from copy import copy
import pickle
from pathlib import Path

import numpy as np
from scipy.spatial.distance import euclidean

import simtk.openmm.app as omma
import simtk.openmm as omm
import simtk.unit as unit

import mdtraj as mdj

from openmm_systems.test_systems import LennardJonesPair

# components for the apparatus
from wepy.resampling.distances.distance import Distance
from wepy.resampling.resamplers.wexplore import WExploreResampler
from wepy.boundary_conditions.unbinding import UnbindingBC
from wepy.runners.openmm import OpenMMRunner, OpenMMState
from wepy.walker import Walker
from wepy.runners.openmm import UNIT_NAMES, GET_STATE_KWARG_DEFAULTS
from wepy.reporter.hdf5 import WepyHDF5Reporter

from wepy.reporter.dashboard import DashboardReporter

from wepy.reporter.receptor.dashboard import ReceptorBCDashboardSection
from wepy.reporter.wexplore.dashboard import WExploreDashboardSection
from wepy.reporter.openmm import OpenMMRunnerDashboardSection

from wepy.reporter.restree import ResTreeReporter
from wepy.util.mdtraj import mdtraj_to_json_topology

from wepy.work_mapper.mapper import Mapper

from wepy.orchestration.orchestrator import Orchestrator
from wepy.orchestration.snapshot import WepySimApparatus
from wepy.orchestration.configuration import Configuration

## PARAMETERS

output_dir = Path("_output")

# Platform used for OpenMM which uses different hardware computation
# kernels. Options are: Reference, CPU, OpenCL, CUDA.

# we use the Reference platform because this is just a test
PLATFORM = 'Reference'

# Langevin Integrator
TEMPERATURE= 300.0*unit.kelvin
FRICTION_COEFFICIENT = 1/unit.picosecond
# step size of time integrations
STEP_SIZE = 0.002*unit.picoseconds

# Resampler parameters

# the maximum weight allowed for a walker
PMAX = 0.5
# the minimum weight allowed for a walker
PMIN = 1e-12

# the maximum number of regions allowed under each parent region
MAX_N_REGIONS = (10, 10, 10, 10)

# the maximum size of regions, new regions will be created if a walker
# is beyond this distance from each voronoi image unless there is an
# already maximal number of regions
MAX_REGION_SIZES = (1, 0.5, .35, .25) # nanometers

# boundary condition parameters

# maximum distance between between any atom of the ligand and any
# other atom of the protein, if the shortest such atom-atom distance
# is larger than this the ligand will be considered unbound and
# restarted in the initial state
CUTOFF_DISTANCE = 1.0 # nm

# reporting parameters

# these are the properties of the states (i.e. from OpenMM) which will
# be saved into the HDF5
SAVE_FIELDS = ('positions', 'box_vectors', 'velocities')
# these are the names of the units which will be stored with each
# field in the HDF5
UNITS = UNIT_NAMES

# orchestration parameters
N_WALKERS = 48


## System and OpenMMRunner

# make the test system
test_sys = LennardJonesPair()

# make the integrator
integrator = omm.LangevinIntegrator(TEMPERATURE, FRICTION_COEFFICIENT, STEP_SIZE)

# make a context and set the positions
context = omm.Context(test_sys.system, copy(integrator))
context.setPositions(test_sys.positions)

# get the data from this context so we have a state to start the
# simulation with
get_state_kwargs = dict(GET_STATE_KWARG_DEFAULTS)
init_sim_state = context.getState(**get_state_kwargs)
init_state = OpenMMState(init_sim_state)

# initialize the runner
runner = OpenMMRunner(test_sys.system, test_sys.topology, integrator, platform=PLATFORM)

## Distance Metric
# we define a simple distance metric for this system, assuming the
# positions are in a 'positions' field
class PairDistance(Distance):

    def __init__(self, metric=euclidean):
        self.metric = metric

    def image(self, state):
        return state['positions']

    def image_distance(self, image_a, image_b):
        dist_a = self.metric(image_a[0], image_a[1])
        dist_b = self.metric(image_b[0], image_b[1])

        return np.abs(dist_a - dist_b)


# make a distance object which can be used to compute the distance
# between two walkers, for our scorer class
distance = PairDistance()

## Resampler
resampler = WExploreResampler(distance=distance,
                              init_state=init_state,
                              max_region_sizes=MAX_REGION_SIZES,
                              max_n_regions=MAX_N_REGIONS,
                              pmin=PMIN, pmax=PMAX)

## Boundary Conditions

# the mdtraj here is needed for the distance function
mdtraj_topology = mdj.Topology.from_openmm(test_sys.topology)

## Reporters if we want them
json_str_top = mdtraj_to_json_topology(mdtraj_topology)

# initialize the unbinding boundary conditions
ubc = UnbindingBC(cutoff_distance=CUTOFF_DISTANCE,
                  initial_state=init_state,
                  topology=json_str_top,
                  ligand_idxs=np.array(test_sys.ligand_indices),
                  receptor_idxs=np.array(test_sys.receptor_indices))

# make a dictionary of units for adding to the HDF5
units = dict(UNIT_NAMES)

hdf5_reporter_kwargs = {'save_fields' : SAVE_FIELDS,
                        'resampler' : resampler,
                        'boundary_conditions' : ubc,
                        'topology' : json_str_top,
                        'units' : units}

wexplore_dash = WExploreDashboardSection(resampler=resampler)
openmm_dash = OpenMMRunnerDashboardSection(runner=runner,
                                           step_time=STEP_SIZE)
ubc_dash = ReceptorBCDashboardSection(bc=ubc)

dashboard_reporter_kwargs = {
    'resampler_dash' : wexplore_dash,
    'runner_dash' : openmm_dash,
    'bc_dash' : ubc_dash,
}

# Resampling Tree
restree_reporter_kwargs = {'resampler' : resampler,
                           'boundary_condition' : ubc,
                           'node_radius' : 3.0,
                           'row_spacing' : 5.0,
                           'step_spacing' : 20.0,
                           'progress_key' : 'min_distances',
                           'max_progress_value' : ubc.cutoff_distance,
                           'colormap_name' : 'plasma'}

# TODO make the restree reporter work

reporter_classes = [
    WepyHDF5Reporter,
    DashboardReporter,
    #ResTreeReporter,
]
reporter_kwargs = [
    hdf5_reporter_kwargs,
    dashboard_reporter_kwargs,
    #restree_reporter_kwargs,
]


sim_apparatus = WepySimApparatus(runner,
                                 resampler=resampler,
                                 boundary_conditions=ubc)

# we also create a default configuration for the orchestrator that
# will be used unless one is given at runtime for the creation of a
# simulation manager
configuration = Configuration(reporter_classes=reporter_classes,
                              reporter_partial_kwargs=reporter_kwargs)

# we also want to set up the orchestrator with some default walkers to
# use to get us started. Otherwise these could be provided from a
# snapshot or on their own. Ideally we only want to have a single
# script setting up an orchestrator and then manage everything else on
# the command line intereactively from then on out
init_weight = 1.0 / N_WALKERS
init_walkers = [Walker(OpenMMState(init_sim_state), init_weight) for i in range(N_WALKERS)]

# then create the seed/root/master orchestrator which will be used
# from here on out
orch = Orchestrator(
    orch_path=str(output_dir / 'LJ-pair.orch.sqlite'),
    mode='w',
)

# set the defaults
orch.set_default_sim_apparatus(sim_apparatus)
orch.set_default_init_walkers(init_walkers)
orch.set_default_configuration(configuration)
orch.gen_default_snapshot()





import os.path as osp
from collections import defaultdict
import itertools as it
import logging

from wepy.resampling.decisions.clone_merge import MultiCloneMergeDecision
from wepy.reporter.dashboard import ResamplerDashboardSection

import numpy as np
import pandas as pd

class REVODashboardSection(ResamplerDashboardSection):

    RESAMPLER_TEMPLATE = \
"""

Resampling Algorithm: {{ name }}

Distance Exponent: {{ dist_exponent }}
Characteristic Distance: {{ char_dist }}
Merge Distance: {{ merge_dist }}

** Resamplig
Cycle index: {{ cycle_idx }}
The percentage of cloned walkers: {{ percentage_cloned_walkers }} %
The percentage of merged walkers: {{ percentage_merged_walkers }} %

** Statistics
Average All to All Distance: {{ avg_distance }}
Minimum All to All Distance: {{ min_distance }}
Maximum All to All Distance: {{ max_distance }}
Variation value = {{ variation }}
"""

    def __init__(self, resampler=None,
                 dist_exponent=None,
                 merge_dist=None,
                 lpmin=None,
                 char_dist=None,
                 seed=None,
                 decision=None,
                 **kwargs):

        if 'name' not in kwargs:
            kwargs['name'] = 'REVOResampler'

        super().__init__(resampler=resampler,
                 dist_exponent=dist_exponent,
                 merge_dist=merge_dist,
                 lpmin=lpmin,
                 char_dist=char_dist,
                 seed=seed,
                 decision=decision,
        )

        if resampler is not None:
            self.dist_exponent = resampler.dist_exponent
            self.merge_dist = resampler.merge_dist
            self.lpmin = resampler.lpmin
            self.char_dist = resampler.char_dist
            self.seed = resampler.seed
            self.decision = resampler.DECISION

        else:
            assert dist_exponent is not None, \
                "if no resampler given must give parameters: dist_exponent"
            assert merge_dist is not None, \
                "if no resampler given must give parameters: merge_dist"
            assert lpmin is not None, \
                "if no resampler given must give parameters: lpmin"
            assert char_dist is not None, \
                "if no resampler given must give parameters: char_dist"
            assert seed is not None, \
                "if no resampler given must give parameters: seed"
            assert decision is not None, \
                "if no resampler given must give parameters: decision"

            self.dist_exponent = dist_exponent
            self.merge_dist = merge_dist
            self.lpmin = lpmin
            self.char_dist = char_dist
            self.seed = seed
            self.decision = decision

        # updatables
        self.percentage_cloned_walkers = 0
        self.percentage_merged_walkers = 0
        #REVO
        self.avg_distance = None
        self.min_distance = None
        self.max_distance = None
        self.variation_values = None

    def update_values(self, **kwargs):

        num_clones = 0
        num_merges = 0
        num_walkers = len(kwargs['resampling_data'])
        for walker_record in kwargs['resampling_data']:
            if walker_record['decision_id'][0]==self.decision.ENUM.CLONE.value:
                num_clones += 1
            elif walker_record['decision_id'][0]==self.decision.ENUM.KEEP_MERGE.value:
                num_merges += 1

        self.percentage_cloned_walkers = (num_clones/num_walkers) * 100
        self.percentage_merged_walkers = (num_merges/num_walkers) * 100

        #Get the statistics
        for resampler_record in kwargs['resampler_data']:
            self.variation_value = resampler_record['variation'][0]
            distance_matrix = resampler_record['distance_matrix']


        #get the upper triangle values of the distance_matrix
        distance_matrix = np.triu(distance_matrix)
        distance_values= distance_matrix[np.where(distance_matrix>0)]
        self.avg_distance = np.average(distance_values)
        self.min_distance = np.min(distance_values)
        self.max_distance  = np.max(distance_values)

        self.cycle_idx = kwargs['cycle_idx']

    def gen_fields(self, **kwargs):

        fields = super().gen_fields(**kwargs)

        new_fields = {
            'dist_exponent' : self.dist_exponent,
            'char_dist' : self.char_dist,
            'merge_dist' : self.merge_dist,
            'cycle_idx' : self.cycle_idx,
            'percentage_cloned_walkers' : self.percentage_cloned_walkers,
            'percentage_merged_walkers' : self.percentage_merged_walkers,
            'avg_distance' : self.avg_distance,
            'min_distance' : self.min_distance,
            'max_distance' : self.max_distance,
            'variation' : self.variation_value
        }

        fields.update(new_fields)

        return fields



from pint import UnitRegistry


from wepy.reporter.dashboard import RunnerDashboardSection

# initialize the unit registry
units = UnitRegistry()

class OpenMMRunnerDashboardSection(RunnerDashboardSection):

    RUNNER_SECTION_TEMPLATE = \
"""

Runner: {{ name }}

Integration Step Size: {{ step_time }}

Single Walker Sampling Time: {{ walker_total_sampling_time }}

Total Sampling Time: {{ total_sampling_time }}
"""

    def __init__(self, runner=None,
                 step_time=None,
                 **kwargs
    ):

        if 'name' not in kwargs:
            kwargs['name'] = 'OpenMMRunner'

        super().__init__(runner=runner,
                         step_time=step_time,
                         **kwargs)

        if runner is None:
            assert step_time is not None, "If no complete runner is given must give parameters: step_time"

            # assume it has units
            self.step_time = step_time

        else:

            simtk_step_time = runner.integrator.getStepSize()
            simtk_val = simtk_step_time.value_in_unit(simtk_step_time.unit)

            # convert to a more general purpose pint unit, which will be
            # used for the dashboards so we don't have the simtk
            # dependency
            self.step_time = simtk_val * units(simtk_step_time.unit.get_name())


        # TODO

        # integrator and params


        # FF and params

        # updatables
        self.walker_total_sampling_time = 0.0 * units('microsecond')
        self.total_sampling_time = 0.0 * units('microsecond')

    def update_values(self, **kwargs):

        super().update_values(**kwargs)

        # amount of new sampling time for each walker
        new_walker_sampling_time = self.step_time * kwargs['n_segment_steps']

        # accumulated sampling time for a single walker
        self.walker_total_sampling_time += new_walker_sampling_time

        # amount of sampling time for all walkers
        new_sampling_time = new_walker_sampling_time * len(kwargs['new_walkers'])

        # accumulated sampling time for the ensemble
        self.total_sampling_time += new_sampling_time


    def gen_fields(self, **kwargs):

        fields = super().gen_fields(**kwargs)

        ## formatting

        # units for the different time scales
        step_size_unit_spec = 'femtosecond'
        sampling_time_unit_spec = 'microsecond'

        # get the unit objects for them
        step_size_unit = units(step_size_unit_spec)
        sampling_time_unit = units(sampling_time_unit_spec)

        # step size string
        step_size_str = str(self.step_time.to(step_size_unit))

        # single walker sampling
        walker_samp_str = str(self.walker_total_sampling_time.to(sampling_time_unit))

        # total sampling
        total_samp_str = str(self.total_sampling_time.to(sampling_time_unit))

        new_fields = {
            'step_time' : step_size_str,
            'walker_total_sampling_time' : walker_samp_str,
            'total_sampling_time' : total_samp_str,
        }

        fields.update(new_fields)

        return fields



import click
import logging
from multiprocessing_logging import install_mp_handler

import simtk.unit as unit

# import all of the sim makers we have available
from wepy_tools.sim_makers.openmm import *

SYSTEM_SIM_MAKERS = {
    'LennardJonesPair' : LennardJonesPairOpenMMSimMaker,
    'LysozymeImplicit' : LysozymeImplicitOpenMMSimMaker,
}

def parse_system_spec(spec):

    sys_spec, runner_platform = spec.split("/")

    runner, platform = runner_platform.split('-')

    return sys_spec, runner, platform

@click.option('-v', '--verbose', is_flag=True)
@click.option('-W', '--work-mapper',
              default='WorkerMapper',
              help="Work mapper for doing work.")
@click.option('-R', '--resampler',
              default='WExplore',
              help="Resampling algorithm.")
@click.argument('n_workers', type=int)
@click.argument('tau', type=float)
@click.argument('n_cycles', type=int)
@click.argument('n_walkers', type=int)
@click.argument('system')
@click.command()
def cli(
        verbose,
        work_mapper,
        resampler,
        n_workers,
        tau,
        n_cycles,
        n_walkers,
        system,
):
    """Run a pre-parametrized wepy simulation.

    \b
    Parameters
    ----------

    \b
    SYSTEM : str
        Which pre-parametrized simulation to run should have the format: System/Runner-Platform

    \b
    N_WALKERS : int
        Number of parallel trajectories to run

    \b
    N_CYCLES : int
        How many cycles to run the simulation for

    \b
    TAU : float
        Cycle simulation time in picoseconds

    \b
    N_WORKERS : int
        Number of worker processes to run on

    \b
    Available Systems
    -----------------

    LennardJonesPair : A pair of Lennard-Jones particles

    LysozymeImplicit : Lysozyme-xylene receptor ligand in implicit solvent (2621 atoms)

    \b
    Available Runners/Platforms
    ---------------------------

    \b
    OpenMM-
      Reference
      CPU
      OpenCL (GPU)
      CUDA (GPU)


    \b
    Available Work Mappers
    ----------------------

    WorkerMapper (default) : parallel python multiprocessing based
                             worker-consumer concurrency model

    WIP not available in test drive yet:

    TaskMapper : parallel python multiprocessing based task-process
                 based concurrency model

    Mapper : non-parallel single-process implementation


    \b
    Available Resamplers
    --------------------

    No : Doesn't do any resampling. Simply runs an ensemble of walkers.

    WExplore : Hierarchical History Dependent Voronoi Binning

    REVO : Stateless and Binless algorithm that rewards in-ensemble novelty.

    \b
    Examples
    --------

    python -m wepy_test_drive LennardJonesPair/OpenMM-CPU 20 10 2 4

    \b
    Notes
    -----

    When using a GPU platform your number of workers should be the
    number of GPUs you want to use.

    """

    if verbose:
        logging.getLogger().setLevel(logging.DEBUG)
        install_mp_handler()
        logging.debug("Starting the test")


    resampler_fullname = resampler + 'Resampler'

    sys_spec, runner, platform = parse_system_spec(system)

    # choose which sim_maker to use
    sim_maker = SYSTEM_SIM_MAKERS[sys_spec]()

    apparatus = sim_maker.make_apparatus(
        platform = platform,
        resampler = resampler_fullname,
    )

    # compute the number of steps to take from tau
    tau = tau * unit.picosecond
    n_steps = round(tau / apparatus.filters[0].integrator.getStepSize())

    config = sim_maker.make_configuration(apparatus,
                                          work_mapper_spec=work_mapper,
                                          platform=platform)

    sim_manager = sim_maker.make_sim_manager(n_walkers, apparatus, config)

    # run the simulation
    sim_manager.run_simulation(n_cycles, n_steps, num_workers=n_workers)


if __name__ == "__main__":

    cli()



"""Module that allows for imposing a kinetically connected network
structure of weighted ensemble simulation data.
"""

from collections import defaultdict
from copy import deepcopy

import numpy as np
import networkx as nx

from wepy.analysis.transitions import transition_counts, counts_d_to_matrix, \
                                      normalize_counts

try:
    import pandas as pd
except ModuleNotFoundError:
    print("Pandas is not installe, that functionality won't work")

class MacroStateNetworkError(Exception):
    """Errors specific to MacroStateNetwork requirements."""
    pass

class BaseMacroStateNetwork():
    """A base class for the MacroStateNetwork which doesn't contain a
    WepyHDF5 object. Useful for serialization of the object and can
    then be reattached later to a WepyHDF5.

    """

    ASSIGNMENTS = 'assignments'
    """Key for the microstates that are assigned to a macrostate."""

    def __init__(self,
                 contig_tree,
                 assg_field_key=None,
                 assignments=None,
                 transition_lag_time=2):
        """Create a network of macrostates from the simulation microstates
        using a field in the trajectory data or precomputed assignments.

        Either 'assg_field_key' or 'assignments' must be given, but not
        both.

        The lag time is default set to 2, which is the natural connection
        between microstates. The lag time can be increased to vary the
        kinetic accuracy of transition probabilities generated through
        Markov State Modelling.

        Alternatively the MacroStateNetwork can also be though of as just
        a way of mapping macrostate properties to the underlying
        microstate data.

        Arguments
        ---------
        contig_tree : ContigTree object

        assg_field_key : str, conditionally optional on 'assignments'
            The field in the WepyHDF5 dataset you want to generate macrostates for.

        assignments : list of list of array_like of dim (n_traj_frames, observable_shape[0], ...),
                      conditionally optional on 'assg_field_key'

            List of assignments for all frames in each run, where each
            element of the outer list is for a run, the elements of
            these lists are lists for each trajectory which are
            arraylikes of shape (n_traj, observable_shape[0], ...).
        """

        self._graph = nx.DiGraph()

        assert not (assg_field_key is None and assignments is None), \
            "either assg_field_key or assignments must be given"

        assert assg_field_key is not None or assignments is not None, \
            "one of assg_field_key or assignments must be given"

        self._base_contig_tree = contig_tree.base_contigtree

        self._assg_field_key = assg_field_key

        # initialize the groups dictionary
        self._node_groups = {}

        # initialize the list of the observables
        self._observables = []

        # initialize the list of available layouts
        self._layouts = []

        # the temporary assignments dictionary
        self._node_assignments = None
        # and temporary raw assignments
        self._assignments = None

        with contig_tree:


            # map the keys to their lists of assignments, depending on
            # whether or not we are using a field from the HDF5 traj or
            # assignments provided separately
            if assg_field_key is not None:
                assert type(assg_field_key) == str, "assignment key must be a string"

                self._key_init(contig_tree)
            else:
                self._assignments_init(assignments)

            # once we have made th dictionary add the nodes to the network
            # and reassign the assignments to the nodes
            self._node_idxs = {}
            for node_idx, assg_item in enumerate(self._node_assignments.items()):
                assg_key, assigs = assg_item
                # count the number of samples (assigs) and use this as a field as well
                num_samples = len(assigs)

                # save the nodes with attributes, we save the node_id
                # as the assg_key, because of certain formats only
                # typing the attributes, and we want to avoid data
                # loss, through these formats (which should be avoided
                # as durable stores of them though)
                self._graph.add_node(assg_key,
                                     node_id=assg_key,
                                     node_idx=node_idx,
                                     assignments=assigs,
                                     num_samples=num_samples)
                self._node_idxs[assg_key] = node_idx

            # also initialize the reverse which is memoized if needed
            self._node_idx_to_id_dict = None

            # now count the transitions between the states and set those
            # as the edges between nodes

            # first get the sliding window transitions from the contig
            # tree, once we set edges for a tree we don't really want to
            # have multiple sets of transitions on the same network so we
            # don't provide the method to add different assignments
            self._transition_lag_time = None
            self._probmat = None
            self._countsmat = None
            if transition_lag_time is not None:


                # get the weights for the walkers so we can compute
                # the weighted transition counts
                weights = [[] for run_idx in contig_tree.wepy_h5.run_idxs]
                for idx_tup, traj_fields_d in contig_tree.wepy_h5.iter_trajs_fields(
                        ['weights'],
                        idxs=True):

                    run_idx, traj_idx = idx_tup

                    weights[run_idx].append(np.ravel(traj_fields_d['weights']))

                # set the lag time attribute
                self._transition_lag_time = transition_lag_time

                # get the transitions
                transitions = []
                for window in contig_tree.sliding_windows(self._transition_lag_time):

                    transition = [window[0], window[-1]]

                    # convert the window trace on the contig to a trace
                    # over the runs
                    transitions.append(transition)

                # then get the counts for those edges
                counts_d = transition_counts(
                    self._assignments,
                    transitions,
                    weights=weights)

                # create the edges and set the counts into them
                for edge, trans_counts in counts_d.items():
                    self._graph.add_edge(*edge, counts=trans_counts)

                # then we also want to get the transition probabilities so
                # we get the counts matrix and compute the probabilities
                # we first have to replace the keys of the counts of the
                # node_ids with the node_idxs
                node_id_to_idx_dict = self.node_id_to_idx_dict()
                self._countsmat = counts_d_to_matrix(
                                    {(node_id_to_idx_dict[edge[0]],
                                      node_id_to_idx_dict[edge[1]]) : counts
                                     for edge, counts in counts_d.items()})

                self._probmat = normalize_counts(self._countsmat)

                # then we add these attributes to the edges in the network
                node_idx_to_id_dict = self.node_id_to_idx_dict()
                for i_id, j_id in self._graph.edges:
                    # i and j are the node idxs so we need to get the
                    # actual node_ids of them
                    i_idx = node_idx_to_id_dict[i_id]
                    j_idx = node_idx_to_id_dict[j_id]

                    # convert to a normal float and set it as an explicitly named attribute
                    self._graph.edges[i_id, j_id]['transition_probability'] = \
                                                                float(self._probmat[i_idx, j_idx])

                    # we also set the general purpose default weight of
                    # the edge to be this.
                    self._graph.edges[i_id, j_id]['Weight'] = \
                                                    float(self._probmat[i_idx, j_idx])


        # then get rid of the assignments dictionary, this information
        # can be accessed from the network
        del self._node_assignments
        del self._assignments

    def _key_init(self, contig_tree):
        """Initialize the assignments structures given the field key to use.

        Parameters
        ----------

        """

        wepy_h5 = contig_tree.wepy_h5

        # blank assignments
        assignments = [[[] for traj_idx in range(wepy_h5.num_run_trajs(run_idx))]
                             for run_idx in wepy_h5.run_idxs]

        test_field = wepy_h5.get_traj_field(
            wepy_h5.run_idxs[0],
            wepy_h5.run_traj_idxs(0)[0],
            self.assg_field_key,
        )

        # WARN: assg_field shapes can come wrapped with an extra
        # dimension. We handle both cases. Test the first traj and see
        # how it is
        unwrap = False

        if len(test_field.shape) == 2 and test_field.shape[1] == 1:
            # then we raise flag to unwrap them
            unwrap = True

        elif len(test_field.shape) == 1 and test_field.shape[0] == 1:
            # then it is unwrapped and we don't need to do anything,
            # just assert the flag to not unwrap
            unwrap = False

        else:

            raise ValueError(f"Wrong shape for an assignment type observable: {test_field.shape}")


        # the raw assignments
        curr_run_idx = -1
        for idx_tup, fields_d in wepy_h5.iter_trajs_fields(
                                         [self.assg_field_key], idxs=True):

            run_idx = idx_tup[0]
            traj_idx = idx_tup[1]


            assg_field = fields_d[self.assg_field_key]

            # if we need to we unwrap the assignements scalar values
            # if they need it
            if unwrap:
                assg_field = np.ravel(assg_field)

            assignments[run_idx][traj_idx].extend(assg_field)

        # then just call the assignments constructor to do it the same
        # way
        self._assignments_init(assignments)

    def _assignments_init(self, assignments):
        """Given the assignments structure sets up the other necessary
        structures.

        Parameters
        ----------
        assignments : list of list of array_like of dim (n_traj_frames, observable_shape[0], ...),
                      conditionally optional on 'assg_field_key'

            List of assignments for all frames in each run, where each
            element of the outer list is for a run, the elements of
            these lists are lists for each trajectory which are
            arraylikes of shape (n_traj, observable_shape[0], ...).

        """

        # set the type for the assignment field
        self._assg_field_type = type(assignments[0])

        # set the raw assignments to the temporary attribute
        self._assignments = assignments

        # this is the dictionary mapping node_id -> the (run_idx, traj_idx, cycle_idx) frames
        self._node_assignments = defaultdict(list)

        for run_idx, run in enumerate(assignments):
            for traj_idx, traj in enumerate(run):
                for frame_idx, assignment in enumerate(traj):
                    self._node_assignments[assignment].append( (run_idx, traj_idx, frame_idx) )

    def node_id_to_idx(self, assg_key):
        """Convert a node_id (which is the assignment value) to a canonical index.

        Parameters
        ----------
        assg_key : node_id

        Returns
        -------
        node_idx : int

        """
        return self.node_id_to_idx_dict()[assg_key]

    def node_idx_to_id(self, node_idx):
        """Convert a node index to its node id.

        Parameters
        ----------
        node_idx : int

        Returns
        -------
        node_id : node_id


        """
        return self.node_idx_to_id_dict()[node_idx]

    def node_id_to_idx_dict(self):
        """Generate a full mapping of node_ids to node_idxs."""
        return self._node_idxs

    def node_idx_to_id_dict(self):
        """Generate a full mapping of node_idxs to node_ids."""


        if self._node_idx_to_id_dict is None:
            rev = {node_idx : node_id for node_id, node_idx in self._node_idxs.items()}
            self._node_idx_to_id_dict = rev
        else:
            rev = self._node_idx_to_id_dict

        # just reverse the dictionary and return
        return rev


    @property
    def graph(self):
        """The networkx.DiGraph of the macrostate network."""
        return self._graph

    @property
    def num_states(self):
        """The number of states in the network."""
        return len(self.graph)

    @property
    def node_ids(self):
        """A list of the node_ids."""
        return list(self.graph.nodes)

    @property
    def contig_tree(self):
        """The underlying ContigTree"""
        return self._base_contig_tree

    @property
    def assg_field_key(self):
        """The string key of the field used to make macro states from the WepyHDF5 dataset.

        Raises
        ------
        MacroStateNetworkError
            If this wasn't used to construct the MacroStateNetwork.

        """
        if self._assg_field_key is None:
            raise MacroStateNetworkError("Assignments were manually defined, no key.")
        else:
            return self._assg_field_key

    @property
    def countsmat(self):
        """Return the transition counts matrix of the network.

        Raises
        ------
        MacroStateNetworkError
            If no lag time was given.

        """

        if self._countsmat is None:
            raise MacroStateNetworkError("transition counts matrix not calculated")
        else:
            return self._countsmat

    @property
    def probmat(self):
        """Return the transition probability matrix of the network.

        Raises
        ------
        MacroStateNetworkError
            If no lag time was given.

        """

        if self._probmat is None:
            raise MacroStateNetworkError("transition probability matrix not set")
        else:
            return self._probmat

    def get_node_attributes(self, node_id):
        """Returns the node attributes of the macrostate.

        Parameters
        ----------
        node_id : node_id

        Returns
        -------
        macrostate_attrs : dict

        """
        return self.graph.nodes[node_id]

    def get_node_attribute(self, node_id, attribute_key):
        """Return the value for a specific node and attribute.

        Parameters
        ----------
        node_id : node_id

        attribute_key : str

        Returns
        -------
        node_attribute

        """
        return self.get_node_attributes(node_id)[attribute_key]

    def get_nodes_attribute(self, attribute_key):
        """Get a dictionary mapping nodes to a specific attribute. """

        nodes_attr = {}
        for node_id in self.graph.nodes:
            nodes_attr[node_id] = self.graph.nodes[node_id][attribute_key]

        return nodes_attr

    def node_assignments(self, node_id):
        """Return the microstates assigned to this macrostate as a run trace.

        Parameters
        ----------
        node_id : node_id

        Returns
        -------
        node_assignments : list of tuples of ints (run_idx, traj_idx, cycle_idx)
            Run trace of the nodes assigned to this macrostate.

        """
        return self.get_node_attribute(node_id, self.ASSIGNMENTS)


    def set_nodes_attribute(self, key, values_dict):
        """Set node attributes for the key and values for each node.

        Parameters
        ----------
        key : str

        values_dict : dict of node_id: values

        """
        for node_id, value in values_dict.items():
            self.graph.nodes[node_id][key] = value

    @property
    def node_groups(self):
        return self._node_groups

    def set_node_group(self, group_name, node_ids):

        # push these values to the nodes themselves, overwriting if
        # necessary
        self._set_group_nodes_attribute(group_name, node_ids)

        # then update the group mapping with this
        self._node_groups[group_name] = node_ids


    def _set_group_nodes_attribute(self, group_name, group_node_ids):

        # the key for the attribute of the group goes in a little
        # namespace prefixed with _group
        group_key = '_groups/{}'.format(group_name)

        # make the mapping
        values_map = {node_id : True if node_id in group_node_ids else False
                      for node_id in self.graph.nodes}

        # then set them
        self.set_nodes_attribute(group_key, values_map)


    @property
    def observables(self):
        """The list of available observables."""
        return self._observables

    def node_observables(self, node_id):
        """Dictionary of observables for each node_id."""

        node_obs = {}
        for obs_name in self.observables:
            obs_key = '_observables/{}'.format(obs_name)
            node_obs[obs_name] = self.get_nodes_attributes(node_id, obs_key)

        return node_obs

    def set_nodes_observable(self, observable_name, node_values):

        # the key for the attribute of the observable goes in a little
        # namespace prefixed with _observable
        observable_key = '_observables/{}'.format(observable_name)

        self.set_nodes_attribute(observable_key, node_values)

        # then add to the list of available observables
        self._observables.append(observable_name)

    @property
    def layouts(self):
        return self._layouts

    def node_layouts(self, node_id):
        """Dictionary of layouts for each node_id."""

        node_layouts = {}
        for layout_name in self.layouts:
            layout_key = '_layouts/{}'.format(layout_name)
            node_layouts[obs_name] = self.get_nodes_attributes(node_id, layout_key)

        return node_layouts

    def set_nodes_layout(self, layout_name, node_values):

        # the key for the attribute of the observable goes in a little
        # namespace prefixed with _observable
        layout_key = '_layouts/{}'.format(layout_name)

        self.set_nodes_attribute(layout_key, node_values)

        # then add to the list of available observables
        if layout_name not in self._layouts:
            self._layouts.append(layout_name)


    def write_gexf(self, filepath, exclude_fields=None, layout=None):
        """Writes a graph file in the gexf format of the network.

        Parameters
        ----------
        filepath : str

        """

        layout_key = None
        if layout is not None:
            layout_key = '_layouts/{}'.format(layout)
            if layout not in self.layouts:
                raise ValueError("Layout not found, use None for no layout")


        if exclude_fields is None:
            exclude_fields = [self.ASSIGNMENTS]
        else:
            exclude_fields.append(self.ASSIGNMENTS)
            exclude_fields = list(set(exclude_fields))

        # exclude the layouts, we will set the viz manually for the layout
        exclude_fields.extend(['_layouts/{}'.format(layout_name)
                               for layout_name in self.layouts])


        # to do this we need to get rid of the assignments in the
        # nodes though since this is not really supported or good to
        # store in a gexf file which is more for visualization as an
        # XML format, so we copy and modify then write the copy
        gexf_graph = deepcopy(self._graph)
        for node in gexf_graph:
            # remove requested fields
            for field in exclude_fields:
                del gexf_graph.nodes[node][field]

            # also remove the fields which are not valid gexf types
            fields = list(gexf_graph.nodes[node].keys())
            for field in fields:
                if type(gexf_graph.nodes[node][field]) not in nx.readwrite.gexf.GEXF.xml_type:
                    del gexf_graph.nodes[node][field]

            if layout_key is not None:

                # set the layout as viz attributes to this
                gexf_graph.nodes[node]['viz'] = self._graph.nodes[node][layout_key]

        nx.write_gexf(gexf_graph, filepath)


    def nodes_to_records(self):

        keys = ['num_samples', 'Weight']

        # add all the groups to the keys
        keys.extend(['_groups/{}'.format(key) for key in self.node_groups.keys()])

        # add the observables
        keys.extend(['_observables/{}'.format(obs) for obs in self.observables])

        recs = []
        for node_id in self.graph.nodes:
            rec = {'node_id' : node_id}
            for key in keys:
                rec[key] = self.get_node_attribute(node_id, key)

            recs.append(rec)

        return recs

    def nodes_to_dataframe(self):
        """Make a dataframe of the nodes and their attributes.

        Not all attributes will be added as they are not relevant to a
        table style representation anyhow.

        The columns will be:

        - node_id
        - node_idx
        - num samples
        - Weight (if available)
        - groups (as booleans) which is anything in the '_groups' namespace
        - observables : anything in the '_observables' namespace and
          will assume to be scalars

        """

        # TODO: set the column order
        # col_order = []

        return pd.DataFrame(self.nodes_to_records())


    def edges_to_records(self):
        """Make a dataframe of the nodes and their attributes.

        Not all attributes will be added as they are not relevant to a
        table style representation anyhow.

        The columns will be:

        - edge_id
        - source
        - target
        - type (directed or undirected)
        - transition counts (if available)
        - transition probability (if available)

        """


        keys = ['counts', 'transition_probability']

        recs = []
        for edge_id in self.graph.edges:
            rec = {'edge_id' : edge_id}
            for key in keys:
                rec[key] = self.graph.edges[edge_id][key]

            recs.append(rec)

        return recs



    def edges_to_dataframe(self):
        """Make a dataframe of the nodes and their attributes.

        Not all attributes will be added as they are not relevant to a
        table style representation anyhow.

        The columns will be:

        - transition counts (if available)
        - transition probability (if available)

        """

        return pd.DataFrame(self.edges_to_records())

    def node_map(self, func, map_func=map):
        """Map a function over the nodes.

        The function should take as its first argument a node_id and
        the second argument a dictionary of the node attributes. This
        will not give access to the underlying trajectory data in the
        HDF5, to do this use the 'node_fields_map' function.

        Extra args not supported use 'functools.partial' to make
        functions with arguments for all data.


        Parameters
        ----------
        func :

        map_func :

        Returns
        -------

        """

        # wrap the function so that we can pass through the node_id
        def func_wrapper(args):
            node_id, node_attrs = args
            return node_id, func(node_attrs)

        # zip the node_ids with the node attributes as an iterator
        node_attr_it = ((node_id,
                         {**self.get_node_attributes(node_id), 'node_id' : node_id})
                        for node_id in self.graph.nodes
                       )

        return {node_id : value for node_id, value
                in map_func(func_wrapper, node_attr_it)}

class MacroStateNetwork():
    """Provides an abstraction over weighted ensemble data in the form of
    a kinetically connected network.

    The MacroStateNetwork refers to any grouping of the so called
    "micro" states that were observed during simulation,
    i.e. trajectory frames, and not necessarily in the usual sense
    used in statistical mechanics. Although it is the perfect vehicle
    for working with such macrostates.

    Because walker trajectories in weighted ensemble there is a
    natural way to generate the edges between the macrostate nodes in
    the network. These edges are determined automatically and a lag
    time can also be specified, which is useful in the creation of
    Markov State Models.
    """


    def __init__(self, contig_tree,
                 base_network=None,
                 assg_field_key=None,
                 assignments=None,
                 transition_lag_time=2):

        self.closed = True
        self._contig_tree = contig_tree
        self._wepy_h5 = self._contig_tree.wepy_h5

        # if we pass a base network use that one instead of building
        # one manually
        if base_network is not None:
            assert isinstance(base_network, BaseMacroStateNetwork)

            self._set_base_network_to_self(base_network)

        else:
            new_network = BaseMacroStateNetwork(contig_tree,
                                                assg_field_key=assg_field_key,
                                                assignments=assignments,
                                                transition_lag_time=transition_lag_time)

            self._set_base_network_to_self(new_network)



    def _set_base_network_to_self(self, base_network):

        self._base_network = base_network

        # then make references to this for the attributes we need

        # attributes
        self._graph = self._base_network._graph
        self._assg_field_key = self._base_network._assg_field_key
        self._node_idxs = self._base_network._node_idxs
        self._node_idx_to_id_dict = self._base_network._node_idx_to_id_dict
        self._transition_lag_time = self._base_network._transition_lag_time
        self._probmat = self._base_network._probmat
        self._countsmat = self._base_network._countsmat

        # functions
        self.node_id_to_idx = self._base_network.node_id_to_idx
        self.node_idx_to_id = self._base_network.node_idx_to_id
        self.node_id_to_idx_dict = self._base_network.node_id_to_idx_dict
        self.node_idx_to_id_dict = self._base_network.node_idx_to_id_dict

        self.get_node_attributes = self._base_network.get_node_attributes
        self.get_node_attribute = self._base_network.get_node_attribute
        self.get_nodes_attribute = self._base_network.get_nodes_attribute
        self.node_assignments = self._base_network.node_assignments
        self.set_nodes_attribute = self._base_network.set_nodes_attribute

        self.node_groups = self._base_network.node_groups
        self.set_node_group = self._base_network.set_node_group
        self._set_group_nodes_attribute = self._base_network._set_group_nodes_attribute
        self.observables = self._base_network.observables
        self.node_observables = self._base_network.node_observables
        self.set_nodes_observable = self._base_network.set_nodes_observable

        self.nodes_to_records =  self._base_network.nodes_to_records
        self.nodes_to_dataframe = self._base_network.nodes_to_dataframe
        self.edges_to_records = self._base_network.edges_to_records
        self.edges_to_dataframe = self._base_network.edges_to_dataframe
        self.node_map = self._base_network.node_map

        self.write_gexf = self._base_network.write_gexf

    def open(self, mode=None):
        if self.closed:
            self.wepy_h5.open(mode=mode)
            self.closed = False
        else:
            raise IOError("This file is already open")

    def close(self):
        self.wepy_h5.close()
        self.closed = True

    def __enter__(self):
        self.wepy_h5.__enter__()
        self.closed = False
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.wepy_h5.__exit__(exc_type, exc_value, exc_tb)
        self.close()


    # from the Base class

    @property
    def graph(self):
        """The networkx.DiGraph of the macrostate network."""
        return self._graph

    @property
    def num_states(self):
        """The number of states in the network."""
        return len(self.graph)

    @property
    def node_ids(self):
        """A list of the node_ids."""
        return list(self.graph.nodes)

    @property
    def assg_field_key(self):
        """The string key of the field used to make macro states from the WepyHDF5 dataset.

        Raises
        ------
        MacroStateNetworkError
            If this wasn't used to construct the MacroStateNetwork.

        """
        if self._assg_field_key is None:
            raise MacroStateNetworkError("Assignments were manually defined, no key.")
        else:
            return self._assg_field_key

    @property
    def countsmat(self):
        """Return the transition counts matrix of the network.

        Raises
        ------
        MacroStateNetworkError
            If no lag time was given.

        """

        if self._countsmat is None:
            raise MacroStateNetworkError("transition counts matrix not calculated")
        else:
            return self._countsmat

    @property
    def probmat(self):
        """Return the transition probability matrix of the network.

        Raises
        ------
        MacroStateNetworkError
            If no lag time was given.

        """

        if self._probmat is None:
            raise MacroStateNetworkError("transition probability matrix not set")
        else:
            return self._probmat






    # unique to the HDF5 holding one
    @property
    def base_network(self):
        return self._base_network

    @property
    def wepy_h5(self):
        """The WepyHDF5 source object for which the contig tree is being constructed. """
        return self._wepy_h5

    def state_to_mdtraj(self, node_id, alt_rep=None):
        """Generate an mdtraj.Trajectory object from a macrostate.

        By default uses the "main_rep" in the WepyHDF5
        object. Alternative representations of the topology can be
        specified.

        Parameters
        ----------
        node_id : node_id

        alt_rep : str
             (Default value = None)

        Returns
        -------
        traj : mdtraj.Trajectory

        """
        with self.wepy_h5:
            return self.wepy_h5.trace_to_mdtraj(self.base_network.node_assignments(node_id),
                                                alt_rep=alt_rep)

    def state_to_traj_fields(self, node_id, alt_rep=None):

        return self.states_to_traj_fields([node_id], alt_rep=alt_rep)

    def states_to_traj_fields(self, node_ids, alt_rep=None):


        node_assignments = []
        for node_id in node_ids:
            node_assignments.extend(self.base_network.node_assignments(node_id))

        with self.wepy_h5:

            # get the right fields
            rep_path = self.wepy_h5._choose_rep_path(alt_rep)
            fields = [rep_path, 'box_vectors']

            return self.wepy_h5.get_trace_fields(node_assignments,
                                                 fields)


    def get_node_fields(self, node_id, fields):
        """Return the trajectory fields for all the microstates in the
        specified macrostate.

        Parameters
        ----------
        node_id : node_id

        fields : list of str
            Field name to retrieve.

        Returns
        -------

        fields : dict of str: array_like
           A dictionary mapping the names of the fields to an array of the field.
           Like fields of a trace.

        """
        node_trace = self.base_network.node_assignments(node_id)

        # use the node_trace to get the weights from the HDF5
        with self.wepy_h5:
            fields_d = self.wepy_h5.get_trace_fields(node_trace, fields)

        return fields_d

    def iter_nodes_fields(self, fields):
        """Iterate over all nodes and return the field values for all the
        microstates for each.

        Parameters
        ----------
        fields : list of str

        Returns
        -------
        nodes_fields : dict of node_id: (dict of field: array_like)
            A dictionary with an entry for each node.
            Each node has it's own dictionary of node fields for each microstate.

        """

        nodes_d = {}
        for node_id in self.graph.nodes:
            fields_d = self.base_network.get_node_fields(node_id, fields)
            nodes_d[node_id] = fields_d

        return nodes_d

    def microstate_weights(self):
        """Returns the weights of each microstate on the basis of macrostates.

        Returns
        -------
        microstate_weights : dict of node_id: ndarray

        """

        node_weights = {}
        for node_id in self.graph.nodes:
            # get the trace of the frames in the node
            node_trace = self.base_network.node_assignments(node_id)

            # use the node_trace to get the weights from the HDF5
            with self.wepy_h5:
                trace_weights = self.wepy_h5.get_trace_fields(node_trace, ['weights'])['weights']

            node_weights[node_id] = trace_weights

        return node_weights

    def macrostate_weights(self):
        """Compute the total weight of each macrostate.

        Returns
        -------
        macrostate_weights : dict of node_id: float

        """

        macrostate_weights = {}
        microstate_weights = self.microstate_weights()
        for node_id, weights in microstate_weights.items():
            macrostate_weights[node_id] = float(sum(weights)[0])

        return macrostate_weights

    def set_macrostate_weights(self):
        """Compute the macrostate weights and set them as node attributes
        'Weight'."""
        self.base_network.set_nodes_attribute('Weight', self.macrostate_weights())

    def node_fields_map(self, func, fields, map_func=map):
        """

        Parameters
        ----------
        func :
            
        fields :
            
        *args :
            
        map_func :
             (Default value = map)
        idxs :
             (Default value = False)
        node_sel :
             (Default value = None)

        Returns
        -------

        """

        # wrap the function so that we can pass through the node_id
        def func_wrapper(args):
            node_id, node_attrs, node_fields = args
            return node_id, func(node_attrs, node_fields)

        # zip the node_ids with the node attributes as an iterator
        node_attr_fields_it = (
            (node_id,
             {**self.get_node_attributes(node_id), 'node_id' : node_id},
             self.get_node_fields(node_id, fields),
            )
                     for node_id in self.graph.nodes)

        return {node_id : value for node_id, value
                in map_func(func_wrapper, node_attr_fields_it)}



import os.path as osp
from copy import deepcopy
import logging
import itertools as it

from wepy.work_mapper.mapper import Mapper, WorkerMapper
from wepy.work_mapper.worker import Worker

class Configuration():
    """ """

    DEFAULT_WORKDIR = osp.realpath(osp.curdir)
    DEFAULT_CONFIG_NAME = "root"
    DEFAULT_NARRATION = ""
    DEFAULT_REPORTER_CLASS = ""

    # if there is to be reporter class in filenames use this template
    # to put it into the filename
    REPORTER_CLASS_SEG_TEMPLATE = ".{}"
    DEFAULT_MODE = 'x'

    def __init__(self,
                 # reporters
                 config_name=None,
                 work_dir=None,
                 mode=None,
                 narration=None,
                 reporter_classes=None,
                 reporter_partial_kwargs=None,
                 # work mappers
                 work_mapper_class=None,
                 work_mapper_partial_kwargs=None,
                 # monitors
                 monitor_class=None,
                 monitor_partial_kwargs=None,
                 # apparatus configuration options
                 apparatus_opts=None,
    ):

        ## reporter stuff

        # reporters and partial kwargs
        if reporter_classes is not None:
            self._reporter_classes = reporter_classes
        else:
            self._reporter_classes = []

        if reporter_partial_kwargs is not None:
            self._reporter_partial_kwargs = reporter_partial_kwargs
        else:
            self._reporter_partial_kwargs = []

        # file path localization variables

        # config string
        if config_name is not None:
            self._config_name = config_name
        else:
            self._config_name = self.DEFAULT_CONFIG_NAME

        if work_dir is not None:
            self._work_dir = work_dir
        else:
            self._work_dir = self.DEFAULT_WORKDIR

        # narration
        if narration is not None:
            narration = "_{}".format(narration) if len(narration) > 0 else ""
            self._narration = narration
        else:
            self._narration = self.DEFAULT_NARRATION


        # file modes, if none are given we set to the default, this
        # needs to be done before generating the reporters
        if mode is not None:
            self._mode = mode
        else:
            self._mode = self.DEFAULT_MODE

        # generate the reporters for this configuration
        self._reporters = self._gen_reporters()

        ## work mapper

        # the partial kwargs that will be passed for reparametrization
        if work_mapper_partial_kwargs is None:
            self._work_mapper_partial_kwargs = {}
        else:
            self._work_mapper_partial_kwargs = work_mapper_partial_kwargs

        # if the number of workers is not given set it to None
        if 'num_workers' not in self._work_mapper_partial_kwargs:
            self._work_mapper_partial_kwargs['num_workers'] = None

        # same for the worker type
        if 'worker_type' not in self._work_mapper_partial_kwargs:
            self._work_mapper_partial_kwargs['worker_type'] = None

        # if the number of workers was sepcified and no work_mapper
        # class was specified default to the WorkerMapper
        if (self._work_mapper_partial_kwargs['num_workers'] is not None) and \
           (work_mapper_class is None):
            self._work_mapper_class = WorkerMapper

        # if no number of workers was specified and no work_mapper
        # class was specified we default to the serial mapper
        elif (self._work_mapper_partial_kwargs['num_workers'] is None) and \
             (work_mapper_class is None):
            self._work_mapper_class = Mapper

        # otherwise if the work_mapper class was given we use it and
        # whatever the number of workers was
        else:
            self._work_mapper_class = work_mapper_class

        # then generate a work mapper
        self._work_mapper = self._work_mapper_class(**self._work_mapper_partial_kwargs)


        ### Monitor options

        # get the names of the reporters in the order they are
        reporter_order = tuple([str(reporter_class.__name__)
                                for reporter_class in self._reporter_classes])

        # init the kwargs for the monitor
        if monitor_partial_kwargs is None:
            self._monitor_partial_kwargs = {}
        else:
            self._monitor_partial_kwargs = monitor_partial_kwargs

        # choose the monitor class (None is okay)
        self._monitor_class = monitor_class

        # generate the object
        if self._monitor_class is not None:

            self._monitor = self._monitor_class(
                reporter_order=reporter_order,
                **self._monitor_partial_kwargs,
            )

        else:
            self._monitor = None

        ### Apparatus options

        # the runtime configuration of the apparatus can be configured
        # via these options.
        self._apparatus_opts = apparatus_opts if apparatus_opts is not None else {}

    @property
    def reporter_classes(self):
        """ """
        return self._reporter_classes

    @property
    def reporter_partial_kwargs(self):
        """ """
        return self._reporter_partial_kwargs

    @property
    def config_name(self):
        """ """
        return self._config_name

    @property
    def work_dir(self):
        """ """
        return self._work_dir

    @property
    def narration(self):
        """ """
        return self._narration

    @property
    def mode(self):
        """ """
        return self._mode

    @property
    def reporters(self):
        """ """
        return self._reporters

    @property
    def work_mapper_class(self):
        """ """
        return self._work_mapper_class

    @property
    def work_mapper_partial_kwargs(self):
        """ """
        return self._work_mapper_partial_kwargs

    @property
    def work_mapper(self):
        """ """
        return self._work_mapper

    @property
    def monitor_class(self):
        """ """
        return self._monitor_class

    @property
    def monitor_partial_kwargs(self):
        """ """
        return self._monitor_partial_kwargs

    @property
    def monitor(self):
        """ """
        return self._monitor


    @property
    def apparatus_opts(self):
        """ """
        return self._apparatus_opts


    def _gen_reporters(self):
        """ """

        # check the extensions of all the reporters. If any of them
        # are the same raise a flag to add the reporter names to the
        # filenames

        # the number of filenames
        all_exts = list(it.chain(*[[ext for ext in rep.SUGGESTED_EXTENSIONS]
                                 for rep in self.reporter_classes]))
        n_exts = len(all_exts)

        # the number of unique ones
        n_unique_exts = len(set(all_exts))

        duplicates = False
        if n_unique_exts < n_exts:
            duplicates = True

        # then go through and make the inputs for each reporter
        reporters = []
        for idx, reporter_class in enumerate(self.reporter_classes):

            # first we have to generate the filenames for all the
            # files this reporter needs. The number of file names the
            # reporter needs is given by the number of suggested
            # extensions it has
            file_paths = []
            for extension in reporter_class.SUGGESTED_EXTENSIONS:


                # if previously found that there are duplicates in the
                # extensions we need to name with the reporter class string
                if duplicates:

                    # use the __name__ attribute of the class and put
                    # it into the template to make a segment out of it
                    reporter_class_seg_str = self.REPORTER_CLASS_SEG_TEMPLATE.format(
                        reporter_class.__name__)

                    # then make the filename with this
                    filename = reporter_class.SUGGESTED_FILENAME_TEMPLATE.format(
                                       narration=self.narration,
                                       config=self.config_name,
                                       reporter_class=reporter_class_seg_str,
                                       ext=extension)

                # otherwise don't use the reporter class names to keep it clean
                else:
                    filename = reporter_class.SUGGESTED_FILENAME_TEMPLATE.format(
                                       narration=self.narration,
                                       config=self.config_name,
                                       reporter_class=self.DEFAULT_REPORTER_CLASS,
                                       ext=extension)


                file_path = osp.join(self.work_dir, filename)

                file_paths.append(file_path)

            modes = [self.mode for i in range(len(file_paths))]

            reporter = reporter_class(file_paths=file_paths, modes=modes,
                                  **self.reporter_partial_kwargs[idx])

            reporters.append(reporter)

        return reporters

    # TODO: remove, not used
    def _gen_work_mapper(self):
        """ """

        work_mapper = self._work_mapper_class(n_workers=self._default_n_workers)

        return work_mapper

    @property
    def reporters(self):
        """ """
        return deepcopy(self._reporters)

    @property
    def work_mapper(self):
        """ """
        return deepcopy(self._work_mapper)

    def reparametrize(self, **kwargs):
        """

        Parameters
        ----------
        **kwargs :
            

        Returns
        -------

        """

        # dictionary of the possible reparametrizations from the
        # current configuration
        params = {

            # related to the work mapper
            'work_mapper_class' : self.work_mapper_class,
            'work_mapper_partial_kwargs' : self.work_mapper_partial_kwargs,

            # monitor
            'monitor_class' : self.monitor_class,
            'monitor_partial_kwargs' : self.monitor_partial_kwargs,

            # those related to the reporters
            'mode' : self.mode,
            'config_name' : self.config_name,
            'work_dir' : self.work_dir,
            'narration' : self.narration,
            'reporter_classes' : self.reporter_classes,
            'reporter_partial_kwargs' : self.reporter_partial_kwargs,

            # apparatus
            'apparatus_opts' : self.apparatus_opts,
        }

        for key, value in kwargs.items():

            # for the partial kwargs we need to update them not
            # completely overwrite
            if key in [
                    'work_mapper_partial_kwargs',
                    'reporter_partial_kwargs',
                    'monitor_partial_kwargs',
            ]:
                if value is not None:
                    params[key].update(value)

            # if the value is given we replace the old one with it
            elif value is not None:
                params[key] = value

        new_configuration = type(self)(**params)

        return new_configuration



"""Glue all the CLIs together into one interface."""

import click

from wepy.orchestration.cli import subgroups as orch_subgroups
from wepy.orchestration.cli import cli as orch_cli

# @click.group()
# def cli():
#     """ """
#     pass

# # add in the sub-clis
# cli.add_command(orch_cli)

# # the orchestrator stuff we keep in the top-level still though
# for subgroup in orch_subgroups:
#     cli.add_command(subgroup)

if __name__ == "__main__":

    orch_cli()



import numpy as np
from scipy.spatial.distance import euclidean

from wepy.runners.openmm import GET_STATE_KWARG_DEFAULTS
from wepy.resampling.distances.distance import Distance
from wepy.boundary_conditions.receptor import UnbindingBC

from openmm_systems.test_systems import LennardJonesPair

from wepy_tools.sim_makers.openmm import OpenMMToolsTestSysSimMaker


## Distance Metric
# we define a simple distance metric for this system, assuming the
# positions are in a 'positions' field
class PairDistance(Distance):

    def __init__(self, metric=euclidean):
        self.metric = metric

    def image(self, state):
        return state['positions']

    def image_distance(self, image_a, image_b):
        dist_a = self.metric(image_a[0], image_a[1])
        dist_b = self.metric(image_b[0], image_b[1])

        return np.abs(dist_a - dist_b)


# class PairUnbinding(BoundaryCondition):

#     pass

class LennardJonesPairOpenMMSimMaker(OpenMMToolsTestSysSimMaker):

    TEST_SYS = LennardJonesPair

    BCS = OpenMMToolsTestSysSimMaker.BCS + [UnbindingBC]

    LIGAND_IDXS = [0]
    RECEPTOR_IDXS = [1]

    UNBINDING_BC_DEFAULTS = {
        'cutoff_distance' : 1.0, # nm
        'periodic' : False,
    }

    DEFAULT_BC_PARAMS = OpenMMToolsTestSysSimMaker.DEFAULT_BC_PARAMS
    DEFAULT_BC_PARAMS.update(
        {
            'UnbindingBC' : UNBINDING_BC_DEFAULTS,
        }
    )

    def make_bc(self, bc_class, bc_params):

        if bc_class == UnbindingBC:
            bc_params.update(
                {
                    'distance' : self.distance,
                    'initial_state' : self.init_state,
                    'topology' : self.json_top(),
                    'ligand_idxs' : self.LIGAND_IDXS,
                    'receptor_idxs' : self.RECEPTOR_IDXS,
                }
            )

        bc = bc_class(**bc_params)

        return bc

    def __init__(self):

        # must set this here since we need it to generate the state,
        # will get called again in the superclass method
        self.getState_kwargs = dict(GET_STATE_KWARG_DEFAULTS)
        if self.GET_STATE_KWARGS is not None:
            self.getState_kwargs.update(self.GET_STATE_KWARGS)

        test_sys = LennardJonesPair()

        init_state = self.make_state(test_sys.system, test_sys.positions)

        super().__init__(
            distance=PairDistance(),
            init_state=init_state,
            system=test_sys.system,
            topology=test_sys.topology,
        )




"""OpenMM molecular dynamics runner with accessory classes.

OpenMM is a library with support for running molecular dynamics
simulations with specific support for fast GPU calculations. The
component based architecture of OpenMM makes it a perfect fit with
wepy.

In addition to the principle OpenMMRunner class there are a few
classes here that make using OpenMM runner more efficient.

First is a WalkerState class (OpenMMState) that wraps the openmm state
object directly, itself is a wrapper around the C++
datastructures. This gives better performance by not performing copies
to a WalkerState dictionary.

Second, is the OpenMMWalker which is identical to the Walker class
except that it enforces the state is an actual instantiation of
OpenMMState. Use of this is optional.

Finally, is the OpenMMGPUWorker class. This is to be used as the
worker type for the WorkerMapper work mapper. This is necessary to
allow passing of the device index to OpenMM for which GPU device to
use.

"""
from copy import copy
import random as rand
from warnings import warn
import logging
import time

from eliot import log_call, start_action

import numpy as np

try:
    import simtk.openmm.app as omma
    import simtk.openmm as omm
    import simtk.unit as unit
except ModuleNotFoundError:
    raise ModuleNotFoundError("OpenMM has not been installed, which this runner requires.")

from wepy.walker import Walker, WalkerState
from wepy.runners.runner import Runner
from wepy.work_mapper.worker import Worker
from wepy.work_mapper.task_mapper import WalkerTaskProcess
from wepy.reporter.reporter import Reporter
from wepy.util.util import box_vectors_to_lengths_angles

## Constants

KEYS = ('positions', 'velocities', 'forces', 'kinetic_energy',
        'potential_energy', 'time', 'box_vectors', 'box_volume',
        'parameters', 'parameter_derivatives')
"""Names of the fields of the OpenMMState."""

# when we use the get_state function from the simulation context we
# can pass options for what kind of data to get, this is the default
# to get all the data. TODO not really sure what the 'groups' keyword
# is for though
GET_STATE_KWARG_DEFAULTS = (('getPositions', True),
                            ('getVelocities', True),
                            ('getForces', True),
                            ('getEnergy', True),
                            ('getParameters', True),
                            ('getParameterDerivatives', True),
                            ('enforcePeriodicBox', True),)
"""Mapping of key word arguments to the simulation.context.getState
method for retrieving data for a simulation state. By default we set
each as True to retrieve all information. The presence or absence of
them is handled by the OpenMMState.

"""

# the Units objects that OpenMM uses internally and are returned from
# simulation data

# TODO: this is never used and we only need the unit names. Its okay
# to use simtk.units here but other runners should use a units sytem
# like pint which is easier to install. So we should remove this since
# its not used.

# UNITS = (('positions_unit', unit.nanometer),
#          ('time_unit', unit.picosecond),
#          ('box_vectors_unit', unit.nanometer),
#          ('velocities_unit', unit.nanometer/unit.picosecond),
#          ('forces_unit', unit.kilojoule / (unit.nanometer * unit.mole)),
#          ('box_volume_unit', unit.nanometer),
#          ('kinetic_energy_unit', unit.kilojoule / unit.mole),
#          ('potential_energy_unit', unit.kilojoule / unit.mole),
#         )
# """Mapping of units identifiers to the corresponding simtk.units Unit objects."""

# the names of the units from the units objects above. This is used
# for saving them to files
UNIT_NAMES = (('positions_unit', unit.nanometer.get_name()),
         ('time_unit', unit.picosecond.get_name()),
         ('box_vectors_unit', unit.nanometer.get_name()),
         ('velocities_unit', (unit.nanometer/unit.picosecond).get_name()),
         ('forces_unit', (unit.kilojoule / (unit.nanometer * unit.mole)).get_name()),
         ('box_volume_unit', unit.nanometer.get_name()),
         ('kinetic_energy_unit', (unit.kilojoule / unit.mole).get_name()),
         ('potential_energy_unit', (unit.kilojoule / unit.mole).get_name()),
        )
"""Mapping of unit identifier strings to the serialized string spec of the unit."""

# a random seed will be chosen from 1 to RAND_SEED_RANGE_MAX when the
# Langevin integrator is created. 0 is the default and special value
# which will then choose a random value when the integrator is created

# TODO: test this isn't needed
# RAND_SEED_RANGE_MAX = 1000000

# the runner for the simulation which runs the actual dynamics
class OpenMMRunner(Runner):
    """Runner for OpenMM simulations."""

    def __init__(self, system, topology, integrator,
                 platform=None,
                 platform_kwargs=None,
                 enforce_box=False):
        """Constructor for OpenMMRunner.

        Parameters
        ----------
        system : simtk.openmm.System object
            The system (forcefields) for the simulation.

        topology : simtk.openmm.app.Topology object
            The topology for you system.

        integrator : subclass simtk.openmm.Integrator object
            Integrator for propagating dynamics.

        platform : str
            The specification for the default computational platform
            to use. Platform can also be set when run_segment is
            called. If None uses OpenMM default platform, see OpenMM
            documentation for all value but typical ones are:
            Reference, CUDA, OpenCL. If value is None the automatic
            platform determining mechanism in OpenMM will be used.

        platform_kwargs : dict of str : bool, optional
            key-values to set for a platform with
            platform.setPropertyDefaultValue as the default for this
            runner.

        enforce_box : bool
            Calls 'context.getState' with 'enforcePeriodicBox' if True.
             (Default value = False)

        Warnings
        --------

        Regarding the 'enforce_box' option.

        When retrieving states from an OpenMM simulation Context, you
        have the option to enforce periodic boundary conditions in the
        resulting atomic positions in a topology aware way that
        doesn't break bonds through boundaries. This is convenient for
        post-processing as this can be a complex task and is not
        readily exposed in the OpenMM API as a standalone function.

        However, in some types of simulations the periodic box vectors
        are ignored (such as implicit solvent ones) despite there
        being no option to not have periodic boundaries in the context
        itself. Likely if you are running one of these kinds of
        simulations you will not pay attention to the box vectors at
        all and the random defaults that exist will be very wrong but
        this incorrectness will not show in a non-wepy simulation with
        openmm unless you are handling the context states
        yourself. Then when you run in wepy the default of True to
        enforce the boxes will be applied and confusingly wrong
        answers will result that are difficult to find root cause of.

        """

        # we save the different components. However, if we are to make
        # this runner picklable we have to convert the SWIG objects to
        # a picklable form
        self.system = system
        self.integrator = integrator

        # these are not SWIG objects
        self.topology = topology
        self.platform_name = platform
        self.platform_kwargs = platform_kwargs

        self.enforce_box = enforce_box

        self.getState_kwargs = dict(GET_STATE_KWARG_DEFAULTS)
        # update with the user based enforce_box
        self.getState_kwargs['enforcePeriodicBox'] = self.enforce_box

        self._cycle_platform = None
        self._cycle_platform_kwargs = None

        # for special monitoring purposes to get split times to debug
        # performance
        self._last_cycle_segments_split_times = []

    @log_call(include_args=[
        'platform',
        'platform_kwargs',
    ],
              include_result=False)
    def pre_cycle(self,
                  platform=None,
                  platform_kwargs=None,
                  **kwargs
    ):

        super().pre_cycle(
            platform=platform,
            platform_kwargs=platform_kwargs,
            **kwargs)

        logging.info(f"Setting the platform ({platform}) in the 'pre_cycle' OpenMM Runner call"
                     f"with platform kwargs: {platform_kwargs}"
        )
        # set the platform and kwargs for this cycle
        self._cycle_platform = platform
        self._cycle_platform_kwargs = platform_kwargs

        # each segment split times will get appended to this
        self._last_cycle_segments_split_times = []


    @log_call(include_args=[],
              include_result=False)
    def post_cycle(self, **kwargs):

        super().post_cycle(**kwargs)

        # remove the platform and kwargs for this cycle
        self._cycle_platform = None
        self._cycle_platform_kwargs = None


    def _resolve_platform(self,
                          platform,
                          platform_kwargs,
        ):
        # resolve which platform to use

        # force usage of environmental one
        if platform is Ellipsis:
            platform_name = None
            platform_kwargs = None

        # use the runtime given one
        elif platform is not None:
            platform_name = platform
            platform_kwargs = platform_kwargs

        # if the pre_cycle configured platform is set use this over
        # the default
        elif self._cycle_platform is not None:
            platform_name = self._cycle_platform
            platform_kwargs = self._cycle_platform_kwargs

        # use the default one
        elif self.platform_name is not None:
            platform_name = self.platform_name
            platform_kwargs = self.platform_kwargs

        # if the default is not set fall back to the environmental one
        else:
            platform_name = None
            platform_kwargs = None

        return platform_name, \
               platform_kwargs,

    @log_call(
        include_args=[
            'segment_length',
            'getState_kwargs',
            'platform',
            'platform_kwargs',
        ],
        include_result=False,
    )
    def run_segment(self,
                    walker,
                    segment_length,
                    getState_kwargs=None,
                    platform=None,
                    platform_kwargs=None,
                    **kwargs):
        """Run dynamics for the walker.

        Parameters
        ----------
        walker : object implementing the Walker interface
            The walker for which dynamics will be propagated.

        segment_length : int or float
            The numerical value that specifies how much dynamics are to be run.

        getState_kwargs : dict of str : bool, optional
            Specify the key-word arguments to pass to
            simulation.context.getState when getting simulation
            states. If None defaults object values.

        platform : str or None or Ellipsis
            The specification for the computational platform to
            use. If None will use the default for the runner and
            ignore platform_kwargs. If Ellipsis forces the use of the
            OpenMM default or environmentally defined platform. See
            OpenMM documentation for all value but typical ones are:
            Reference, CUDA, OpenCL. If value is None the automatic
            platform determining mechanism in OpenMM will be used.

        platform_kwargs : dict of str : bool, optional
            key-values to set for a platform with
            platform.setPropertyDefaultValue for this segment only.


        Returns
        -------
        new_walker : object implementing the Walker interface
            Walker after dynamics was run, only the state should be modified.

        """

        run_segment_start = time.time()

        # set the kwargs that will be passed to getState
        tmp_getState_kwargs = getState_kwargs

        logging.info("Default 'getState_kwargs' in runner: "
                     f"{self.getState_kwargs}")

        logging.info("'getState_kwargs' passed to 'run_segment' : "
                     f"{getState_kwargs}")

        # start with the object value
        getState_kwargs = copy(self.getState_kwargs)
        if tmp_getState_kwargs is not None:
            getState_kwargs.update(tmp_getState_kwargs)

        logging.info("After resolving 'getState_kwargs' that will be used are: "
                     f"{getState_kwargs}")


        gen_sim_start = time.time()

        # make a copy of the integrator for this particular segment
        new_integrator = copy(self.integrator)
        # force setting of random seed to 0, which is a special
        # value that forces the integrator to choose another
        # random number
        new_integrator.setRandomNumberSeed(0)

        ## Platform

        logging.info("Default 'platform' in runner: "
                     f"{self.platform_name}")

        logging.info("pre_cycle set 'platform' in runner: "
                     f"{self._cycle_platform}")

        logging.info("'platform' passed to 'run_segment' : "
                     f"{platform}")

        logging.info("Default 'platform_kwargs' in runner: "
                     f"{self.platform_kwargs}")

        logging.info("pre_cycle set 'platform_kwargs' in runner: "
                     f"{self._cycle_platform_kwargs}")


        logging.info("'platform_kwargs' passed to 'run_segment' : "
                     f"{platform_kwargs}")

        platform_name, platform_kwargs = self._resolve_platform(
            platform, platform_kwargs
        )


        logging.info("Resolved 'platform' : "
                     f"{platform_name}")

        logging.info("Resolved 'platform_kwargs' : "
                     f"{platform_kwargs}")



        # create simulation object

        ## create the platform and customize



        # if a platform was given we use it to make a Simulation object
        if platform_name is not None:

            logging.info("Using platform configured in code.")

            # get the platform by its name to use
            platform = omm.Platform.getPlatformByName(platform_name)
            logging.info(f"Platform object created: {platform}")

            if platform_kwargs is None:
                platform_kwargs = {}

            # set properties from the kwargs if they apply to the platform
            for key, value in platform_kwargs.items():

                if key in platform.getPropertyNames():

                    logging.info(f"Setting platform property: {key} : {value}")
                    platform.setPropertyDefaultValue(key, value)

                else:
                    warn(f"Platform kwargs given ({key} : {value}) "
                         f"but is not valid for this platform ({platform_name})")

            # make a new simulation object
            simulation = omma.Simulation(self.topology, self.system,
                                         new_integrator, platform)

        # otherwise just use the default or environmentally defined one
        else:
            logging.info("Using environmental platform.")
            simulation = omma.Simulation(self.topology, self.system,
                                         new_integrator)

        # set the state to the context from the walker
        simulation.context.setState(walker.state.sim_state)

        gen_sim_end = time.time()
        gen_sim_time = gen_sim_end - gen_sim_start

        logging.info("Time to generate the system: {}".format(gen_sim_time))


        # actually run the simulation

        steps_start = time.time()

        # Run the simulation segment for the number of time steps
        with start_action(action_type="OpenMM Simulation.steps") as ommsim_cx:
            simulation.step(segment_length)

        steps_end = time.time()
        steps_time = steps_end - steps_start


        logging.info("Time to run {} sim steps: {}".format(segment_length, steps_time))


        get_state_start = time.time()


        get_state_end = time.time()
        get_state_time = get_state_end - get_state_start
        logging.info("Getting context state time: {}".format(get_state_time))


        # generate the new state/walker
        new_state = self.generate_state(simulation, segment_length,
                                        walker, getState_kwargs)

        # create a new walker for this
        new_walker = OpenMMWalker(new_state, walker.weight)

        run_segment_end = time.time()
        run_segment_time = run_segment_end - run_segment_start
        logging.info("Total internal run_segment time: {}".format(run_segment_time))


        segment_split_times = {
            'gen_sim_time' :  gen_sim_time,
            'steps_time' : steps_time,
            'get_state_time' : get_state_time,
            'run_segment_time' : run_segment_time,
        }

        self._last_cycle_segments_split_times.append(segment_split_times)

        return new_walker

    @log_call(include_args=['getState_kwargs'],
              include_result=False)
    def generate_state(self, simulation, segment_length, starting_walker, getState_kwargs):
        """Method for generating a wepy compliant state from an OpenMM
        simulation object and data about the last segment of dynamics run.

        Parameters
        ----------

        simulation : simtk.openmm.app.Simulation object
            A complete simulation object from which the state will be extracted.

        segment_length : int
            The number of integration steps run in a segment of simulation.

        starting_walker : wepy.walker.Walker subclass object
            The walker that was the beginning of this segment of simyulation.

        getState_kwargs : dict of str : bool
            Specify the key-word arguments to pass to
            simulation.context.getState when getting simulation
            states.

        Returns
        -------

        new_state : wepy.runners.openmm.OpenMMState object
            A new state from the simulation state.

        This method is meant to be called from within the
        `run_segment` method during a simulation. It can be customized
        in subclasses to allow for the addition of custom attributes
        for a state, in addition to the base ones implemented in the
        interface to the openmm simulation state in OpenMMState.

        The extra arguments to this function are data that would allow
        for the calculation of integral values over the duration of
        the segment, such as time elapsed and differences from the
        starting state.

        """

        # save the state of the system with all possible values
        new_sim_state = simulation.context.getState(**getState_kwargs)

        # make an OpenMMState wrapper with this
        new_state = OpenMMState(new_sim_state)

        return new_state


class OpenMMState(WalkerState):
    """Walker state that wraps an simtk.openmm.State object.

    The keys for which values in the state are available are given by
    the KEYS module constant (accessible through the class constant of
    the same name as well).

    Additional fields can be added to these states through passing
    extra kwargs to the constructor. These will be automatically given
    a suffix of "_OTHER" to avoid name clashes.

    """

    KEYS = KEYS
    """The provided attribute keys for the state."""

    OTHER_KEY_TEMPLATE = "{}_OTHER"
    """String formatting template for attributes not set in KEYS."""

    def __init__(self, sim_state, **kwargs):
        """Constructor for OpenMMState.

        Parameters
        ----------
        state : simtk.openmm.State object
            The simulation state retrieved from the simulation constant.

        kwargs : optional

            Additional attributes to set for the state. Will add the
        "_OTHER" suffix to the keys

        """

        # save the simulation state
        self._sim_state = sim_state

        # save additional data if given
        self._data = {}
        for key, value in kwargs.items():

            # if the key is already in the sim_state keys we need to
            # modify it and raise a warning
            if key in self.KEYS:

                warn("Key {} in kwargs is already taken by this class, renaming to {}".format(
                    self.OTHER_KEY_TEMPLATE).format(key))

                # make a new key
                new_key = self.OTHER_KEY_TEMPLATE.format(key)

                # set it in the data
                self._data[new_key] = value

            # otherwise just set it
            else:
                self._data[key] = value

    @property
    def sim_state(self):
        """The underlying simtk.openmm.State object this is wrapping."""
        return self._sim_state

    def __getitem__(self, key):

        # if this was a key for data not mapped from the OpenMM.State
        # object we use the _data attribute
        if (key not in self.KEYS) and (
                (not key.startswith('parameters')) and
                (not key.startswith('parameter_derivatives'))
        ):
            return self._data[key]

        # otherwise we have to specifically get the correct data and
        # process it into an array from the OpenMM.State
        else:

            if key == 'positions':
                return self.positions_values()
            elif key == 'velocities':
                return self.velocities_values()
            elif key == 'forces':
                return self.forces_values()
            elif key == 'kinetic_energy':
                return self.kinetic_energy_value()
            elif key == 'potential_energy':
                return self.potential_energy_value()
            elif key == 'time':
                return self.time_value()
            elif key == 'box_vectors':
                return self.box_vectors_values()
            elif key == 'box_volume':
                return self.box_volume_value()

            # handle the parameters differently since they are dictionaries of values
            elif key.startswith('parameters'):
                parameters_dict = self.parameters_values()
                if parameters_dict is None:
                    return None
                else:

                    # TODO: this was an attempt at a general way to do
                    # this but it doesn't work and I only ever need
                    # one nested level, so for now we just implement it that way
                    # return self._get_nested_attr_from_compound_key(key, parameters_dict)

                    param_key = key.split('/')[-1]
                    return parameters_dict[param_key]

            elif key.startswith('parameter_derivatives'):
                pd_dict = self.parameter_derivatives_values()
                if pd_dict is None:
                    return None
                else:
                    return self._get_nested_attr_from_compound_key(key, pd_dict)

    ## Array properties

    # Positions
    @property
    def positions(self):
        """The positions of the state as a numpy array simtk.units.Quantity object."""
        try:
            return self.sim_state.getPositions(asNumpy=True)
        except:
            warn("Unknown exception handled from `self.sim_state.getPositions()`, "
                     "this is probably because this attribute is not in the State.")
            return None

    @property
    def positions_unit(self):
        """The units (as a simtk.units.Unit object) the positions are in."""
        return self.positions.unit

    def positions_values(self):
        """The positions of the state as a numpy array in the positions_unit
        simtk.units.Unit. This is what is returned by the __getitem__
        accessor.

        """
        return self.positions.value_in_unit(self.positions_unit)

    # Velocities
    @property
    def velocities(self):
        """The velocities of the state as a numpy array simtk.units.Quantity object."""
        try:
            return self.sim_state.getVelocities(asNumpy=True)
        except:
            warn("Unknown exception handled from `self.sim_state.getVelocities()`, "
                     "this is probably because this attribute is not in the State.")
            return None

    @property
    def velocities_unit(self):
        """The units (as a simtk.units.Unit object) the velocities are in."""
        return self.velocities.unit

    def velocities_values(self):
        """The velocities of the state as a numpy array in the velocities_unit
        simtk.units.Unit. This is what is returned by the __getitem__
        accessor.

        """

        velocities = self.velocities
        if velocities is None:
            return None
        else:
            return self.velocities.value_in_unit(self.velocities_unit)

    # Forces
    @property
    def forces(self):
        """The forces of the state as a numpy array simtk.units.Quantity object."""
        try:
            return self.sim_state.getForces(asNumpy=True)
        except:
            warn("Unknown exception handled from `self.sim_state.getForces()`, "
                     "this is probably because this attribute is not in the State.")
            return None

    @property
    def forces_unit(self):
        """The units (as a simtk.units.Unit object) the forces are in."""
        return self.forces.unit

    def forces_values(self):
        """The forces of the state as a numpy array in the forces_unit
        simtk.units.Unit. This is what is returned by the __getitem__
        accessor.

        """

        forces = self.forces
        if forces is None:
            return None
        else:
            return self.forces.value_in_unit(self.forces_unit)

    # Box Vectors
    @property
    def box_vectors(self):
        """The box vectors of the state as a numpy array simtk.units.Quantity object."""
        try:
            return self.sim_state.getPeriodicBoxVectors(asNumpy=True)
        except:
            warn("Unknown exception handled from `self.sim_state.getPeriodicBoxVectors()`, "
                     "this is probably because this attribute is not in the State.")
            return None

    @property
    def box_vectors_unit(self):
        """The units (as a simtk.units.Unit object) the box vectors are in."""
        return self.box_vectors.unit

    def box_vectors_values(self):
        """The box vectors of the state as a numpy array in the
        box_vectors_unit simtk.units.Unit. This is what is returned by
        the __getitem__ accessor.

        """

        box_vectors = self.box_vectors
        if box_vectors is None:
            return None
        else:
            return self.box_vectors.value_in_unit(self.box_vectors_unit)


    ## non-array properties

    # Kinetic Energy
    @property
    def kinetic_energy(self):
        """The kinetic energy of the state as a numpy array simtk.units.Quantity object."""
        try:
            return self.sim_state.getKineticEnergy()
        except:
            warn("Unknown exception handled from `self.sim_state.getKineticEnergy()`, "
                     "this is probably because this attribute is not in the State.")
            return None

    @property
    def kinetic_energy_unit(self):
        """The units (as a simtk.units.Unit object) the kinetic energy is in."""
        return self.kinetic_energy.unit

    def kinetic_energy_value(self):
        """The kinetic energy of the state as a numpy array in the kinetic_energy_unit
        simtk.units.Unit. This is what is returned by the __getitem__
        accessor.

        """

        kinetic_energy = self.kinetic_energy
        if kinetic_energy is None:
            return None
        else:
            return np.array([self.kinetic_energy.value_in_unit(self.kinetic_energy_unit)])

    # Potential Energy
    @property
    def potential_energy(self):
        """The potential energy of the state as a numpy array simtk.units.Quantity object."""
        try:
            return self.sim_state.getPotentialEnergy()
        except:
            warn("Unknown exception handled from `self.sim_state.getPotentialEnergy()`, "
                     "this is probably because this attribute is not in the State.")
            return None

    @property
    def potential_energy_unit(self):
        """The units (as a simtk.units.Unit object) the potential energy is in."""
        return self.potential_energy.unit

    def potential_energy_value(self):
        """The potential energy of the state as a numpy array in the potential_energy_unit
        simtk.units.Unit. This is what is returned by the __getitem__
        accessor.

        """

        potential_energy = self.potential_energy
        if potential_energy is None:
            return None
        else:
            return np.array([self.potential_energy.value_in_unit(self.potential_energy_unit)])

    # Time
    @property
    def time(self):
        """The time of the state as a numpy array simtk.units.Quantity object."""
        try:
            return self.sim_state.getTime()
        except:
            warn("Unknown exception handled from `self.sim_state.getTime()`, "
                     "this is probably because this attribute is not in the State.")
            return None

    @property
    def time_unit(self):
        """The units (as a simtk.units.Unit object) the time is in."""
        return self.time.unit

    def time_value(self):
        """The time of the state as a numpy array in the time_unit
        simtk.units.Unit. This is what is returned by the __getitem__
        accessor.

        """

        time = self.time
        if time is None:
            return None
        else:
            return np.array([self.time.value_in_unit(self.time_unit)])

    # Box Volume
    @property
    def box_volume(self):
        """The box volume of the state as a numpy array simtk.units.Quantity object."""
        try:
            return self.sim_state.getPeriodicBoxVolume()
        except:
            warn("Unknown exception handled from `self.sim_state.getPeriodicBoxVolume()`, "
                     "this is probably because this attribute is not in the State.")
            return None

    @property
    def box_volume_unit(self):
        """The units (as a simtk.units.Unit object) the box volume is in."""
        return self.box_volume.unit

    def box_volume_value(self):
        """The box volume of the state as a numpy array in the box_volume_unit
        simtk.units.Unit. This is what is returned by the __getitem__
        accessor.

        """

        box_volume = self.box_volume
        if box_volume is None:
            return None
        else:
            return np.array([self.box_volume.value_in_unit(self.box_volume_unit)])

    ## Dictionary properties
    ## Unitless

    # Parameters
    @property
    def parameters(self):
        """The parameters of the state as a dictionary mapping the names of
        the parameters to their values which are numpy array
        simtk.units.Quantity objects.

        """

        try:
            return self.sim_state.getParameters()
        except:
            warn("Unknown exception handled from `self.sim_state.getParameters()`, "
                     "this is probably because this attribute is not in the State.")
            return None

    @property
    def parameters_unit(self):
        """The units for each parameter as a dictionary mapping parameter
        names to their corresponding unit as a simtk.units.Unit
        object.

        """
        param_units = {key : None for key, val in self.parameters.items()}
        return param_units

    def parameters_values(self):
        """The parameters of the state as a dictionary mapping the name of the
        parameter to a numpy array in the unit for the parameter of the
        same name in the parameters_unit corresponding
        simtk.units.Unit object. This is what is returned by the
        __getitem__ accessor using the compound key syntax with the
        prefix 'parameters', e.g. state['parameter/paramA'] for the
        parameter 'paramA'.

        """

        if self.parameters is None:
            return None

        param_arrs = {key : np.array(val) for key, val
                          in self.parameters.items()}

        # return None if there is nothing in this
        if len(param_arrs) == 0:
            return None
        else:
            return param_arrs

    # Parameter Derivatives
    @property
    def parameter_derivatives(self):
        """The parameter derivatives of the state as a dictionary mapping the
        names of the parameters to their values which are numpy array
        simtk.units.Quantity objects.

        """

        try:
            return self.sim_state.getEnergyParameterDerivatives()
        except:
            warn("Unknown exception handled from `self.sim_state.getEnergyParameterDerivatives()`, "
                     "this is probably because this attribute is not in the State.")
            return None

    @property
    def parameter_derivatives_unit(self):
        """The units for each parameter derivative as a dictionary mapping
        parameter names to their corresponding unit as a
        simtk.units.Unit object.

        """

        param_units = {key : None for key, val in self.parameter_derivatives.items()}
        return param_units

    def parameter_derivatives_values(self):
        """The parameter derivatives of the state as a dictionary mapping the
        name of the parameter to a numpy array in the unit for the
        parameter of the same name in the parameters_unit
        corresponding simtk.units.Unit object. This is what is
        returned by the __getitem__ accessor using the compound key
        syntax with the prefix 'parameter_derivatives',
        e.g. state['parameter_derivatives/paramA'] for the parameter
        'paramA'.

        """

        if self.parameter_derivatives is None:
            return None

        param_arrs = {key : np.array(val) for key, val
                          in self.parameter_derivatives.items()}

        # return None if there is nothing in this
        if len(param_arrs) == 0:
            return None
        else:
            return param_arrs

    # for the dict attributes we need to transform the keys for making
    # a proper state where all __getitem__ things are arrays
    def _dict_attr_to_compound_key_dict(self, root_key, attr_dict):
        """Transform a dictionary of values within the compound key 'root_key'
        to a dictionary mapping compound keys to values.

        For example give the root_key 'parameters' and the parameters
        dictionary {'paramA' : 1.234} returns {'parameters/paramA' : 1.234}.

        Parameters
        ----------
        root_key : str
            The compound key prefix
        attr_dict : dict of str : value
            The dictionary with simple keys within the root key namespace.

        Returns
        -------
        compound_key_dict : dict of str : value
            The dictionary with the compound keys.

        """

        key_template = "{}/{}"
        cmpd_key_d = {}
        for key, value in attr_dict.items():

            new_key = key_template.format(root_key, key)
            # if this is a proper feature
            if type(value) == np.ndarray:
                cmpd_key_d[new_key] = value
            elif hasattr(value, '__getitem__'):
                cmpd_key_d.update(self._dict_attr_to_compound_key_dict(new_key, value))
            else:
                raise TypeError("Unsupported attribute type")

        return cmpd_key_d

    def _get_nested_attr_from_compound_key(self, compound_key, compound_feat_dict):
        """Get arbitrarily deeply nested compound keys from the full
        dictionary tree.

        Parameters
        ----------
        compound_key : str
            Compound key separated by '/' characters

        compound_feat_dict : dict
            Dictionary of arbitrary depth

        Returns
        -------
        value
            Value requested by the key.

        """

        key_components = compound_key.split('/')

        # if there is only one component of the key then it is not
        # really compound, we won't complain just return the
        # "dictionary" if it is not actually a dict like
        if not hasattr(compound_feat_dict, '__getitem__'):
            raise TypeError("Must provide a dict-like with the compound key")

        value = compound_feat_dict[key_components[0]]

        # if the value itself is compound recursively fetch the value
        if hasattr(value, '__getitem__') and len(key_components[1:]) > 0:

            subgroup_key = '/'.join(key_components[1:])

            return self._get_nested_attr_from_compound_key(subgroup_key, value)

        elif hasattr(value, '__getitem__') and len(key_components[1:]) < 1:
            raise ValueError("Key does not reference a leaf node of attribute")

        # otherwise we have the right key so return the object
        else:
            return value

    def parameters_features(self):
        """Returns a dictionary of the parameters with their appropriate
        compound keys. This can be used for placing them in the same namespace
        as the rest of the attributes."""

        parameters = self.parameters_values()
        if parameters is None:
            return None
        else:
            return self._dict_attr_to_compound_key_dict('parameters', parameters)

    def parameter_derivatives_features(self):
        """Returns a dictionary of the parameter derivatives with their appropriate
        compound keys. This can be used for placing them in the same namespace
        as the rest of the attributes."""

        parameter_derivatives = self.parameter_derivatives_values()
        if parameter_derivatives is None:
            return None
        else:
            return self._dict_attr_to_compound_key_dict('parameter_derivatives',
                                                        parameter_derivatives)

    def omm_state_dict(self):
        """Return a dictionary with all of the default keys from the wrapped
        simtk.openmm.State object"""

        feature_d = {'positions' : self.positions_values(),
                'velocities' : self.velocities_values(),
                'forces' : self.forces_values(),
                'kinetic_energy' : self.kinetic_energy_value(),
                'potential_energy' : self.potential_energy_value(),
                'time' : self.time_value(),
                'box_vectors' : self.box_vectors_values(),
                'box_volume' : self.box_volume_value(),
                    }

        params = self.parameters_features()
        param_derivs = self.parameter_derivatives_features()
        if params is not None:
            feature_d.update(params)
        if param_derivs is not None:
            feature_d.update(param_derivs)

        return feature_d

    def dict(self):
        # documented in superclass

        d = {}
        for key, value in self._data.items():
            d[key] = value
        for key, value in self.omm_state_dict().items():
            d[key] = value
        return d

    def to_mdtraj(self, topology):
        """Returns an mdtraj.Trajectory object from this walker's state.

        Parameters
        ----------
        topology : mdtraj.Topology object
            Topology for the state.

        Returns
        -------
        state_traj : mdtraj.Trajectory object

        """

        import mdtraj as mdj
        # resize the time to a 1D vector
        unitcell_lengths, unitcell_angles = box_vectors_to_lengths_angles(self.box_vectors)
        return mdj.Trajectory(np.array([self.positions_values()]),
                              unitcell_lengths=[unitcell_lengths],
                              unitcell_angles=[unitcell_angles],
                              topology=topology)

def gen_sim_state(positions, system, integrator,
                  getState_kwargs=None):
    """Convenience function for generating an omm.State object.

    Parameters
    ----------

    positions : arraylike of float
        The positions for the system you want to set

    system : openmm.app.System object

    integrator : openmm.Integrator object

    Returns
    -------

    sim_state : openmm.State object

    """

    # handle the getState_kwargs
    tmp_getState_kwargs = getState_kwargs

    # start with the defaults
    getState_kwargs = dict(GET_STATE_KWARG_DEFAULTS)

    # if there were customizations use them
    if tmp_getState_kwargs is not None:
        getState_kwargs.update(tmp_getState_kwargs)

    # generate a throwaway context, using the reference platform so we
    # don't screw up other platform stuff later in the same process
    platform = omm.Platform.getPlatformByName('Reference')
    context = omm.Context(system, copy(integrator), platform)

    # set the positions
    context.setPositions(positions)

    # then just retrieve it as a state using the default kwargs
    sim_state = context.getState(**getState_kwargs)

    return sim_state

def gen_walker_state(positions, system, integrator,
                     getState_kwargs=None):
    """Convenience function for generating a wepy walker State object for
    an openmm simulation state.

    Parameters
    ----------

    positions : arraylike of float
        The positions for the system you want to set

    system : openmm.app.System object

    integrator : openmm.Integrator object

    Returns
    -------

    walker_state : wepy.runners.openmm.OpenMMState object

    """

    state = OpenMMState(gen_sim_state(positions, system, integrator,
                                      getState_kwargs=getState_kwargs))

    return state


class OpenMMWalker(Walker):
    """Walker for OpenMMRunner simulations.

    This simply enforces the use of an OpenMMState object for the
    walker state attribute.

    """

    def __init__(self, state, weight):
        # documented in superclass

        assert isinstance(state, OpenMMState), \
            "state must be an instance of class OpenMMState not {}".format(type(state))

        super().__init__(state, weight)


class OpenMMCPUWorker(Worker):
    """Worker for OpenMM GPU simulations (CUDA or OpenCL platforms).

    This is intended to be used with the wepy.work_mapper.WorkerMapper
    work mapper class.

    This class must be used in order to ensure OpenMM runs jobs on the
    appropriate GPU device.

    """

    NAME_TEMPLATE = "OpenMMCPUWorker-{}"
    """The name template the worker processes are named to substituting in
    the process number."""

    DEFAULT_NUM_THREADS = 1

    def __init__(self, *args, **kwargs):

        if 'num_threads' not in kwargs:
            num_threads = self.DEFAULT_NUM_THREADS
        else:
            num_threads = kwargs.pop('num_threads')

        super().__init__(*args,
                         num_threads=num_threads,
                         **kwargs)

    def run_task(self, task):
        # documented in superclass

        # make the platform kwargs dictionary
        platform_options = {'Threads' : str(self.attributes['num_threads'])}

        # run the task and pass in the DeviceIndex for OpenMM to
        # assign work to the correct GPU
        return task(platform_kwargs=platform_options)


class OpenMMGPUWorker(Worker):
    """Worker for OpenMM GPU simulations (CUDA or OpenCL platforms).

    This is intended to be used with the wepy.work_mapper.WorkerMapper
    work mapper class.

    This class must be used in order to ensure OpenMM runs jobs on the
    appropriate GPU device.

    """

    NAME_TEMPLATE = "OpenMMGPUWorker-{}"
    """The name template the worker processes are named to substituting in
    the process number."""

    def run_task(self, task):

        # get the platform
        platform = self.mapper_attributes['platform']

        # get the device index from the attributes
        device_id = self.mapper_attributes['device_ids'][self._worker_idx]

        # make the platform kwargs dictionary
        platform_options = {'DeviceIndex' : str(device_id)}

        logging.info(f"platform={platform}, platform_options={platform_options}")

        return task(
            platform=platform,
            platform_kwargs=platform_options,
        )


class OpenMMCPUWalkerTaskProcess(WalkerTaskProcess):

    NAME_TEMPLATE = "OpenMM_CPU_Walker_Task-{}"


    def run_task(self, task):

        if 'num_threads' in self.mapper_attributes:
            num_threads = self.mapper_attributes['num_threads']

            # make the platform kwargs dictionary
            platform_options = {'Threads' : str(num_threads)}

            logging.info(f"Threads={num_threads}")

        else:
            platform_options = {}

        return task(
            platform_kwargs=platform_options,
        )

class OpenMMGPUWalkerTaskProcess(WalkerTaskProcess):

    NAME_TEMPLATE = "OpenMM_GPU_Walker_Task-{}"


    def run_task(self, task):


        logging.info(f"Starting to run a task as worker {self._worker_idx}")

        # get the platform
        platform = self.mapper_attributes['platform']

        # get the device index from the attributes
        device_id = self.mapper_attributes['device_ids'][self._worker_idx]

        # make the platform kwargs dictionary
        platform_options = {'DeviceIndex' : str(device_id)}

        logging.info(f"platform={platform}, platform_options={platform_options}")

        return task(
            platform=platform,
            platform_kwargs=platform_options,
        )



import multiprocessing as mp
import queue as pyq
import time
import logging
from warnings import warn
import sys
import traceback
import signal
import pickle

from wepy.work_mapper.mapper import ABCWorkerMapper, WrapperException, TaskException, Task

from wepy.walker import Walker

class TaskProcessException(WrapperException):
    pass

class TaskProcessKilledError(ChildProcessError):
    pass

class TaskMapper(ABCWorkerMapper):
    """Process-per-task mapper.

    This method of work mapper starts new processes for each runner
    segment task that needs to be run. This allows cheap copying of
    shared state using the operating system primitives. On linux this
    would be either 'fork' (default) or 'spawn'. Fork is cheap but
    doesn't initialize certain process namespace things, whereas spawn
    is much more expensive but properly cleans things up. Fork should
    be sufficient in most cases, however spawn may be needed when you
    have some special contexts in the parent process. This is the case
    with starting CUDA contexts in the main parent process and then
    forking new processes from it. We suggest using fork and avoiding
    making these kinds of contexts in the main process.

    This method avoids using shared memory or sending objects through
    interprocess communication (that has a serialization and
    deserialization cost associated with them) by using OS copying
    mechanism. However, a new process will be created each cycle for
    each walker in the simulation. So if you want a large number of
    walkers you may experience a large overhead. If your walker states
    are very small or a very fast serializer is available you may also
    not benefit from full process address space copies. Instead the
    WorkerMapper may be better suited.

    """

    def __init__(self,
                 walker_task_type=None,
                 num_workers=None,
                 segment_func=None,
                 **kwargs):

        super().__init__(num_workers=num_workers,
                         segment_func=segment_func,
                         **kwargs)

        # choose the type of the worker
        if walker_task_type is None:
            self._walker_task_type = WalkerTaskProcess
            warn("walker_task_type not given using the default base class")
            logging.warn("walker_task_type not given using the default base class")
        else:
            self._walker_task_type = walker_task_type

        # initialize a list to put results in
        self.results = None

        # this is meant to be a transient variable, will be initialized and deinitialized
        self._walker_processes = None

    def init(self, **kwargs):

        super().init(**kwargs)

        # now that we have started the processes register the handler
        # for SIGTERM signals that will clean up our children cleanly
        signal.signal(signal.SIGTERM, self._sigterm_shutdown)

    def _sigterm_shutdown(self, signum, frame):

        logging.critical("Received external SIGTERM, forcing shutdown.")

        self.force_shutdown()

        logging.critical("Shutdown complete.")



    @property
    def walker_task_type(self):
        """The callable that generates a worker object.

        Typically this is just the type from the class definition of
        the Worker where the constructor is called.

        """
        return self._walker_task_type


    def force_shutdown(self):

        # send sigterm signals to processes to kill them
        for walker_idx, walker_process in enumerate(self._walker_processes):

            logging.critical("Sending SIGTERM message on {} to worker {}".format(
                self._irq_parent_conns[walker_idx].fileno(), walker_idx))

            # send a kill message to the worker
            self._irq_parent_conns[walker_idx].send(signal.SIGTERM)

        logging.critical("All kill messages sent to workers")


        # wait for the walkers to finish and handle errors in them
        # appropriately
        alive_walkers = [walker.is_alive() for walker in self._walker_processes]
        walker_exitcodes = {}
        premature_exit = False
        while any(alive_walkers):

            for walker_idx, walker in enumerate(self._walker_processes):

                if not alive_walkers[walker_idx]:
                    continue

                if walker.is_alive():
                    pass

                # otherwise the walker is done
                else:
                    alive_walkers[walker_idx] = False
                    walker_exitcodes[walker_idx] = walker.exitcode



    def map(self, *args, **kwargs):

        # run computations in a Manager context
        with self._mp_ctx.Manager() as manager:

            num_walkers = len(args[0])

            # to manage access to worker resources we use a queue with
            # the index of the worker
            worker_queue = manager.Queue()

            # put the workers onto the queue
            for worker_idx in range(self.num_workers):
                worker_queue.put(worker_idx)

            # initialize segment times for workers to
            # fill in
            worker_segment_times = manager.dict()

            # initialize for the number of workers, since these will be
            # the slots to put timing results in
            for i in range(self.num_workers):
                worker_segment_times[i] = []

            # make a shared list for the walker results
            results = manager.list()

            # since this will be indexed by walker index initialize the
            # length of the array
            for walker in range(num_walkers):
                results.append(None)


            # use pipes for communication channels between this parent
            # process and the children for sending specific interrupts
            # such as the signal to kill them. Note that the clean way to
            # end the process is to send poison pills on the task queue,
            # this is for other stuff. IRQ is a common abbreviation for
            # interrupts
            self._irq_parent_conns = []

            # unpack the generator for the kwargs
            kwargs = {key : list(kwarg) for key, kwarg in kwargs.items()}

            # create the task based processes
            self._walker_processes = []
            for walker_idx, task_args in enumerate(zip(*args)):

                task_kwargs = {key : value[walker_idx] for key, value in kwargs.items()}

                # make the interrupt pipe
                parent_conn, child_conn = self._mp_ctx.Pipe()
                self._irq_parent_conns.append(parent_conn)

                # start a process for this walker
                walker_process = self.walker_task_type(walker_idx,
                                                       self._attributes,
                                                       self._func,
                                                       task_args,
                                                       task_kwargs,
                                                       worker_queue,
                                                       results,
                                                       worker_segment_times,
                                                       child_conn,
                )

                walker_process.start()

                self._walker_processes.append(walker_process)

            new_walkers = [None for _ in range(num_walkers)]
            results_found = [False for _ in range(num_walkers)]
            while not all(results_found):

                # go through the results list and handle the values that may be there
                for walker_idx, result in enumerate(results):

                    if results_found[walker_idx]:
                        continue

                    # logging.info("Checking for walker {}".format(walker_idx))

                    # first check to see if any of the task processes were
                    # terminated from the system
                    if self._irq_parent_conns[walker_idx].poll():

                        irq = self._irq_parent_conns[walker_idx].recv()

                        if issubclass(type(irq), TaskProcessKilledError):

                            # just terminate if a worker goes down. We
                            # could handle this better but it is not implemented now
                            logging.critical(
                                "Process {} was killed by sigterm, shutting down.".format(
                                walker_process[walker_idx].name))

                            logging.info(
                                "Recovery is possible here, but is not implemented "
                                "so we opt to fail fast and let you know a problem exists."
                                "Please use checkpointing to avoid lost data.")

                            self.force_shutdown()
                            logging.critical("Shutdown complete.")

                        logging.debug("Received {} acknowledgement from {}".format(ack,
                                                                                   worker.name))

                    # if no interrupts were handled we continue

                    # if it is None no response has been made at all
                    # yet, this is the initialized value
                    if result is None:
                        pass


                    # walker results are returned serialized as
                    # pickles, they are packed into a tuple so that we
                    # can associate them with an explicit marker, if
                    # we have a tuple then we can handle that
                    # appropriately
                    elif type(result) == tuple:

                        logging.debug("Received a results tuple")

                        assert len(result) == 2, "Result tuples should be only be (ID, pickle)"

                        result_id, payload = result

                        # there was a walker successfully returned
                        if result_id == 'Walker':

                            logging.debug("Received a serialized results walker")

                            # deserialize
                            logging.debug("deserializing")
                            new_walker = pickle.loads(payload)

                            logging.info("Got result for walker {}".format(walker_idx))

                            new_walkers[walker_idx] = new_walker
                            results_found[walker_idx] = True

                        else:
                            raise ValueError("Unkown result ID: {}".format(result_id))

                    elif issubclass(type(result), TaskException):

                        logging.critical(
                            "Exception encountered in a task which is unrecoverable."
                            "You will need to reconfigure your components in a stable manner.")

                        self.force_shutdown()

                        logging.critical("Shutdown complete.")
                        raise result

                    elif issubclass(type(result), TaskProcessException):

                        # we make just an error message to say that errors
                        # in the worker may be due to the network or
                        # something and could recover
                        logging.error("Exception encountered in the work mapper task process."
                                      "Recovery possible, see further messages.")

                        # However, the current implementation doesn't
                        # support retries or whatever so we issue a
                        # critical log informing that it has been elevated
                        # to critical and will force shutdown
                        logging.critical(
                            "Task process error mode resiliency not supported at this time."
                            "Performing force shutdown and simulation ending.")

                        self.force_shutdown()

                        logging.critical("Shutdown complete.")
                        raise result

                    elif issubclass(type(result), Exception):
                        logging.critical("Unknown exception {} encountered.".format(result))

                        self.force_shutdown()

                        logging.critical("Shutdown complete.")

                        raise result

                    else:
                        logging.critical("Unknown result value {} encountered.".format(result))

                        self.force_shutdown()

                        logging.critical("Shutdown complete.")


            # save the managed list of the recorded worker times locally
            for key, val in worker_segment_times.items():
                self._worker_segment_times[key] = val

            # wait for the processes to end
            # for walker in self._walker_processes:
            #     walker.join()
            #     logging.info("Joined {}".format(walker.name))

        # deinitialize the current walker processes
        self._walker_processes = None

        return new_walkers


class WalkerTaskProcess(mp.Process):

    NAME_TEMPLATE = "Walker-{}"

    def __init__(self,
                 walker_idx,
                 mapper_attributes,
                 func,
                 task_args,
                 task_kwargs,
                 worker_queue,
                 results_list,
                 worker_segment_times,
                 interrupt_connection,
                 **kwargs
                 ):

        # initialize the process customizing the name
        mp.Process.__init__(self, name=self.NAME_TEMPLATE.format(walker_idx), **kwargs)

        # the idea with this TaskProcess thing is that we pass in all
        # the data to the constructor to create a "thunk" (a closure
        # that is ready to be run without arguments) and then when run
        # is called there will be no arguments to be passed. This
        # simplifies the flow of data and underscores that the task is
        # the process.

        # task arguments
        self._func = func
        self._task_args = task_args
        self._task_kwargs = task_kwargs


        self.walker_idx = walker_idx
        self._worker_idx = None
        self.mapper_attributes = mapper_attributes

        # set the managed datastructure proxies as an attribute so we
        self._worker_queue = worker_queue
        self._results_list = results_list
        self._worker_segment_times = worker_segment_times
        self._irq_channel = interrupt_connection

        # also register the SIGTERM signal handler for graceful
        # shutdown with reporting to mapper
        signal.signal(signal.SIGTERM, self._external_sigterm_shutdown)


    def _external_sigterm_shutdown(self, signum, frame):

        logging.debug("Received external SIGTERM kill command.")

        logging.debug("Alerting mapper that this will be honored.")

        # send an error to the mapper that the worker has been killed
        self._irq_channel.send(TaskProcessKilledError(
            "{} (pid: {}) killed by external SIGTERM signal".format(self.name, self.pid)))

        logging.debug("Acknowledgment sent")

        logging.debug("Shutting down process")


    def _shutdown(self):
        """The normal shutdown which can be ordered by the work mapper."""

        logging.debug("Received SIGTERM kill command from mapper")

        logging.debug("Acknowledging kill request will be honored")

        # report back that we are shutting down with a True
        self._irq_channel.send(True)

        logging.debug("Acknowledgment sent")

        logging.debug("Shutting down process")



    @property
    def attributes(self, key):
        return self._attributes

    @attributes.getter
    def attributes(self, key):
        return self._attributes[key]

    def _run_task(self, task):

        # run the task thunk
        logging.info("{}: Running task".format(self.name))
        try:
            result = self.run_task(task)
        except Exception as task_exception:

            # get the traceback for the exception
            tb = sys.exc_info()[2]

            msg = "Exception '{}({})' caught in a task.".format(
                                   type(task_exception).__name__, task_exception)
            traceback_log_msg = \
                """Traceback:
--------------------------------------------------------------------------------
{}
--------------------------------------------------------------------------------
                """.format(''.join(traceback.format_exception(
                    type(task_exception), task_exception, tb)),
                )

            logging.critical("{}: ".format(self.name) + msg + '\n' + traceback_log_msg)

            # raise a TaskException to distinguish it from the worker
            # errors with the metadata about the original exception

            raise TaskException("Error occured during task execution, recovery not possible.",
                            wrapped_exception=task_exception,
                            tb=tb)

        return result

    def run_task(self, task):

        logging.info("Running an unspecialized task")

        return task()

    def run(self):

        logging.debug("{}: starting to run".format(self.name))

        # try to run the worker and it's task, except either class of
        # error that can come from it either from the worker
        # (WorkerException) or the task (TaskException) and communicate it
        # back to the main process

        # if we get an exception there is some cleanup logic
        run_exception = None

        try:

            # run the worker, which will retrieve its task from the
            # queue attempt to run the task, and if it succeeds will
            # put the results on the result queue, if the task fails
            # it will catch it and wrap it as a task exception
            self._run_walker()

        except TaskException as task_exception:

            logging.error("{}: TaskException caught".format(self.name))

            run_exception = task_exception

        # anything else is considered a WorkerException so take the
        # original exception and generate a worker exception from that
        except Exception as exception:

            logging.debug("{}: TaskProcessException Error caught".format(self.name))


            # get the traceback
            tb = sys.exc_info()[2]

            msg = "Exception '{}({})' caught in a task process.".format(
                                   type(exception).__name__, exception)
            traceback_log_msg = \
                """Traceback:
--------------------------------------------------------------------------------
{}
--------------------------------------------------------------------------------
                """.format(''.join(traceback.format_exception(
                    type(exception), exception, tb)),
                )

            logging.error(msg + '\n' + traceback_log_msg)

            # raise a TaskError to distinguish it from the worker
            # errors with the metadata about the original exception

            walker_exception = TaskProcessException(
                "Error occured during task process execution.",
                wrapped_exception=exception,
                tb=tb)

            run_exception = walker_exception

        # raise worker_exception
        if run_exception is not None:

            logging.debug("{}: Putting exception in managed results list".format(self.name))

            # then put the exception and the traceback onto the queue
            # so we can communicate back to the parent process
            try:
                self._results_list[self.walker_idx] = run_exception
            except BrokenPipeError as exc:
                logging.error(
                    "{}: Pipe is broken indicating the root process has already exited:\n{}".format(
                        self.name, exc))

    def _run_walker(self):

        logging.info("Walker process started as name: {}; PID: {}".format(self.name,
                                                                          self.pid))

        # lock a worker, then pop it off the queue so no other process
        # tries to use it
        worker_received = False
        while not worker_received:

            # pop off a worker to use it, this will block until it
            # receives a worker
            try:
                worker_idx = self._worker_queue.get_nowait()
            except pyq.Empty:
                pass

            # always do on a successful get
            else:

                if type(worker_idx) == int:
                    worker_received = True
                    logging.info("{}: acquired worker {}".format(self.name,
                                                                 worker_idx))

                # if it is a shutdown signal we do so
                elif worker_idx is signal.SIGTERM:
                    logging.info(
                        "{}: SIGTERM signal received from mapper. Shutting down.".format(
                        self.name))

                    self._shutdown()

                    return None


            # check to see if there is any signals on the interrupt channel
            if self._irq_channel.poll():

                # get the message
                message = self._irq_channel.recv()

                logging.debug("{}: Received message from mapper on filehandle {}: {}".format(
                    self.name, self._irq_channel.fileno(), message))

                # handle the message


                # check for signals to die
                if message is signal.SIGTERM:

                    logging.critical(
                        f"{self.name}: SIGTERM signal received from mapper. Shutting down."
                    )

                    self._shutdown()

                    return None

                else:
                    logging.error("{}: Message not recognized, continuing operations and"
                                  " sending error to mapper".format(self.name))
                    self._irq_channel.send(
                        ValueError(
                            "Message: {} not recognized continuing operations".format(
                                message)))

        # after the wait loop we can now perform our work

        self._worker_idx = worker_idx

        # generate the task thunk
        task = Task(self._func, *self._task_args, **self._task_kwargs)

        # run the task
        start = time.time()
        logging.info("{}: running function. Time: {}".format(self.name,
                                                      time.time()))

        # run the task doing the proper handling of the task
        # exception, this can raise a task exception
        result = self._run_task(task)

        logging.info("{}: finished running function. Time {}".format(self.name,
                                                                     time.time()))
        end = time.time()

        # we separately set the result to the results list. This is so
        # we can debug performance problems, and dissect what is due
        # to computation time and what is due to communication
        logging.info("{}: Setting value to results list. Time {}".format(self.name,
                                                                     time.time()))

        logging.debug("Serializing the result walker")

        serial_result = pickle.dumps(result)
        logging.debug("Putting tagged serialized walker tuple on managed results list")
        self._results_list[self.walker_idx] = ('Walker', serial_result)

        logging.info("{}: Finished setting value to results list. Time {}".format(self.name,
                                                                                  time.time()))

        # put the worker back onto the queue since we are done using it
        self._worker_queue.put(worker_idx)

        logging.info("{}: released worker {}".format(self.name,
                                                     worker_idx))

        # add the time for this segment to the collection of the worker times
        segment_time = end - start
        seg_times = self._worker_segment_times[worker_idx] + [segment_time]

        # we must explicitly set the new value in total to trigger an
        # update of the real dictionary. In place modification of
        # proxy objects has no effect
        self._worker_segment_times[worker_idx] = seg_times

        logging.info("{}: Exiting normally having completed the task".format(self.name))



"""WIP: Reporter that produces a text file that gives high level
information on the progress of a simulation.

"""
from collections import defaultdict
import itertools as it
import logging
from datetime import datetime
import time
from copy import copy

import numpy as np
import pandas as pd
from tabulate import tabulate
from jinja2 import Template

import logging

from wepy.reporter.reporter import ProgressiveFileReporter

class DashboardReporter(ProgressiveFileReporter):
    """A text based report of the status of a wepy simulation.

    This serves as a container for different dashboard components to
    go inside.

    """

    FILE_ORDER = ("dashboard_path",)
    SUGGESTED_EXTENSIONS = ("dash.org",)

    # TODO: add in a section for showing the number of walkers in each
    # cycle. This isn't relevant for our constant walker number
    # simulations though so I have elided it following YAGNI

    SIMULATION_SECTION_TEMPLATE = \
"""
Init Datetime: {{ init_date_time }}
Last write Datetime: {{ curr_date_time }}

Total Run time: {{ total_run_time }} s

Last Cycle Index: {{ last_cycle_idx }}
Number of Cycles: {{ n_cycles }}

** Walkers Summary

{{ walker_cycle_summary_table }}

"""

    PERFORMANCE_SECTION_TEMPLATE =\
"""
Average Cycle Time: {{ avg_cycle_time }}

{% if avg_runner_time %}Average Runner Time: {{ avg_runner_time }}{% else %}{% endif %}
{% if avg_bc_time %}Average Boundary Conditions Time: {{ avg_bc_time }}{% else %}{% endif %}
{% if avg_resampling_time %}Average Resampling Time: {{ avg_resampling_time }}{% else %}{% endif %}


Worker Avg. Segment Times:

{{ worker_avg_segment_time }}

** Cycle Performance Log

{{ cycle_log }}

** Worker Performance Log

{{ performance_log }}


"""

    DASHBOARD_TEMPLATE = \
"""* Simulation

{{ simulation }}

{% if resampler %}* Resampler{% else %}{% endif %}
{% if resampler %}{{ resampler }}{% else %}{% endif %}

{% if boundary_condition %}* Boundary Condition{% else %}{% endif %}
{% if boundary_condition %}{{ boundary_condition }}{% else %}{% endif %}

{% if runner %}* Runner{% else %}{% endif %}
{% if runner %}{{ runner }}{% else %}{% endif %}

* Performance

{{ performance }}

"""


    def __init__(self,
                 resampler_dash=None,
                 runner_dash=None,
                 bc_dash=None,
                 **kwargs):
        """

        Parameters
        ----------

        resampler_dash

        runner_dash

        bc_dash
        """

        super().__init__(**kwargs)

        self.resampler_dash = resampler_dash
        self.runner_dash = runner_dash
        self.bc_dash = bc_dash

        # recalculated values

        # general simulation

        # total number of cycles run
        self.n_cycles = 0

        # start date and time
        self.init_date_time = None
        self.init_sys_time = None

        # the total run time
        self.total_run_time = None

        # walker probabilities statistics
        self.walker_prob_summaries = []


        # performance
        self.cycle_compute_times = []
        self.cycle_runner_times = []
        self.cycle_bc_times = []
        self.cycle_resampling_times = []
        self.worker_records = []


    def init(self, **kwargs):

        super().init(**kwargs)

        self.init_date_time = datetime.today()
        self.total_run_time = self.init_date_time

        self.init_sys_time = time.time()

    def calc_walker_summary(self, **kwargs):

        walker_weights = [walker.weight for walker in kwargs['new_walkers']]

        summary = {
            'total' : np.sum(walker_weights),
            'min' : np.min(walker_weights),
            'max' : np.max(walker_weights),
        }

        return summary

    def update_values(self, **kwargs):

        ### simulation

        self.n_cycles += 1
        self.walker_prob_summaries.append(self.calc_walker_summary(**kwargs))

        self.update_performance_values(**kwargs)

        # update all the sections values
        if self.resampler_dash is not None:
            self.resampler_dash.update_values(**kwargs)
        if self.runner_dash is not None:
            self.runner_dash.update_values(**kwargs)
        if self.bc_dash is not None:
            self.bc_dash.update_values(**kwargs)

    def update_performance_values(self, **kwargs):

        ## worker specific performance

        # only do this part if there were any workers
        if len(kwargs['worker_segment_times']) > 0:

            # log of segment times for workers
            for worker_idx, segment_times in kwargs['worker_segment_times'].items():
                for segment_time in segment_times:
                    record = (
                        kwargs['cycle_idx'],
                        kwargs['n_segment_steps'],
                        worker_idx,
                        segment_time,
                    )
                    self.worker_records.append(record)

            # make a table out of these and compute the averages for each
            # worker
            worker_df = pd.DataFrame(self.worker_records, columns=('cycle_idx', 'n_steps',
                                                                   'worker_idx', 'segment_time'))
            # the aggregated table for the workers
            self.worker_agg_table = worker_df.groupby('worker_idx')\
                                    [['segment_time']].aggregate(np.mean)
            self.worker_agg_table.rename(columns={'segment_time' : 'avg_segment_time (s)'},
                                         inplace=True)

        else:
            self.worker_records = []
            self.worker_agg_table = pd.DataFrame({'avg_segment_time (s)' : []})


        ## cycle times

        # log of the components times
        self.cycle_runner_times.append(kwargs['cycle_runner_time'])
        self.cycle_bc_times.append(kwargs['cycle_bc_time'])
        self.cycle_resampling_times.append(kwargs['cycle_resampling_time'])

        # TODO: produces nan if one of them is not given
        # add up the three components to get the overall cycle time
        cycle_time = (kwargs['cycle_runner_time'] +
                      kwargs['cycle_bc_time'] +
                      kwargs['cycle_resampling_time'])

        # log of cycle times
        self.cycle_compute_times.append(cycle_time)

        # average of cycle components times
        self.avg_runner_time = np.mean(self.cycle_runner_times)
        self.avg_bc_time = np.mean(self.cycle_bc_times)
        self.avg_resampling_time = np.mean(self.cycle_resampling_times)

        # average cycle time
        self.avg_cycle_time = np.mean(self.cycle_compute_times)



    def write_dashboard(self, report_str):
        """Write the dashboard to the file."""

        with open(self.file_path, mode=self.mode) as dashboard_file:
            dashboard_file.write(report_str)

    def gen_sim_section(self, **kwargs):
        """"""

        walker_df = pd.DataFrame(self.walker_prob_summaries)
        walker_summary_tbl_str = tabulate(walker_df,
                                          headers=walker_df.columns,
                                          tablefmt='orgtbl')

        # render the simulation section
        sim_section_d = {
            'init_date_time' : self.init_date_time,
            'curr_date_time' : datetime.today().isoformat(),
            'total_run_time' : time.time() - self.init_sys_time,
            'last_cycle_idx' : kwargs['cycle_idx'],
            'n_cycles' : self.n_cycles,
            'walker_cycle_summary_table' : walker_summary_tbl_str,
        }

        sim_section_str = Template(self.SIMULATION_SECTION_TEMPLATE).render(
            **sim_section_d
        )

        return sim_section_str

    def gen_performance_section(self, **kwargs):

        # log of cycle times
        cycle_table_colnames = ('cycle_time (s)',
                                'runner_time (s)',
                                'boundary_conditions_time (s)',
                                'resampling_time (s)')

        cycle_table_df = pd.DataFrame({'cycle_times (s)' : self.cycle_compute_times,
                                       'runner_time (s)' : self.cycle_runner_times,
                                       'boundary_conditions_time (s)' : self.cycle_bc_times,
                                       'resampling_time (s)' : self.cycle_resampling_times},
                                      columns=cycle_table_colnames)

        cycle_table_str = tabulate(cycle_table_df,
                                   headers=cycle_table_df.columns,
                                   tablefmt='orgtbl')

        # log of workers performance
        worker_table_colnames = ('cycle_idx', 'n_steps', 'worker_idx', 'segment_time (s)',)
        worker_table_df = pd.DataFrame(self.worker_records, columns=worker_table_colnames)
        worker_table_str = tabulate(worker_table_df,
                                    headers=worker_table_df.columns,
                                    tablefmt='orgtbl',
                                    showindex=False)


        # table for aggregeated worker stats
        worker_agg_table_str = tabulate(self.worker_agg_table,
                                        headers=self.worker_agg_table.columns,
                                        tablefmt='orgtbl')


        performance_section_d = {
            'avg_cycle_time' : self.avg_cycle_time,
            'worker_avg_segment_time' : worker_agg_table_str,
            'cycle_log' : cycle_table_str,
            'performance_log' : worker_table_str,

            # optionals
            'avg_runner_time' : self.avg_runner_time,
            'avg_bc_time' : self.avg_bc_time,
            'avg_resampling_time' : self.avg_resampling_time,
        }

        performance_section_str = Template(self.PERFORMANCE_SECTION_TEMPLATE).render(
            **performance_section_d
        )

        return performance_section_str

    def report(self, **kwargs):

        # update the values that update each call to report
        self.update_values(**kwargs)

        # the two sections that are always there
        sim_section_str = self.gen_sim_section(**kwargs)
        performance_section_str = self.gen_performance_section(**kwargs)

        # the other optional sections

        # resampler
        if self.resampler_dash is not None:
            resampler_section_str = self.resampler_dash.gen_resampler_section(**kwargs)
        else:
            resampler_section_str = None

        # runner
        if self.runner_dash is not None:
            runner_section_str = self.runner_dash.gen_runner_section(**kwargs)
        else:
            runner_section_str = None

        # boundary conditions
        if self.bc_dash is not None:
            bc_section_str = self.bc_dash.gen_bc_section(**kwargs)
        else:
            bc_section_str = None

        # render the whole template
        report_str = Template(self.DASHBOARD_TEMPLATE).render(
            simulation=sim_section_str,
            resampler=resampler_section_str,
            boundary_condition=bc_section_str,
            runner=runner_section_str,
            performance=performance_section_str
        )

        # write the thing
        self.write_dashboard(report_str)



class ResamplerDashboardSection():

    RESAMPLER_SECTION_TEMPLATE = \
"""
Resampler: {{ name }}
"""

    def __init__(self, resampler=None,
                 name=None,
                 **kwargs):

        if resampler is not None:
            self.resampler_name = type(resampler).__name__

        elif name is not None:
            self.resampler_name = name

        else:
            self.resampler_name = "Unknown"

    def update_values(self, **kwargs):
        pass

    def gen_fields(self, **kwargs):

        fields = {
            'name' : self.resampler_name
        }

        return fields


    def gen_resampler_section(self, **kwargs):

        section_kwargs = self.gen_fields(**kwargs)

        section_str = Template(self.RESAMPLER_SECTION_TEMPLATE).render(
            **section_kwargs
        )

        return section_str

class RunnerDashboardSection():

    RUNNER_SECTION_TEMPLATE = \
"""
Runner: {{ name }}
"""

    def __init__(self, runner=None,
                 name=None,
                 **kwargs):

        if runner is not None:
            self.runner_name = type(runner).__name__

        elif name is not None:
            self.runner_name = name

        else:
            self.runner_name = "Unknown"

    def update_values(self, **kwargs):
        pass

    def gen_fields(self, **kwargs):

        fields = {
            'name' : self.runner_name
        }

        return fields


    def gen_runner_section(self, **kwargs):

        section_kwargs = self.gen_fields(**kwargs)


        section_str = Template(self.RUNNER_SECTION_TEMPLATE).render(
            **section_kwargs
        )

        return section_str


class BCDashboardSection():

    BC_SECTION_TEMPLATE = \
"""

Boundary Condition: {{ name }}

Total Number of Dynamics segments: {{ total_n_walker_segments }}

Total Number of Crossings: {{ total_crossings }}

Cumulative Boundary Crossed Weight: {{ total_crossed_weight }}

** Warping Log

{{ warping_log }}

"""


    WARP_RECORD_COLNAMES = (
        'cycle_idx',
        'walker_idx',
        'weight',
        'target_idx',
        'discontinuous',
    )


    def __init__(self, bc=None,
                 discontinuities=None,
                 name=None,
                 **kwargs):

        if bc is not None:
            self.bc_name = type(bc).__name__

        elif name is not None:
            self.bc_name = name

        else:
            self.bc_name = "Unknown"


        if bc is not None:
            self.bc_discontinuities = copy(bc.DISCONTINUITY_TARGET_IDXS)

        else:
            assert discontinuities is not None, \
                "If the bc is not given must give parameter: discontinuities"
            self.bc_discontinuities = discontinuities

        self.warp_records = []
        self.total_n_walker_segments = 0
        self.total_crossings = 0
        self.total_crossed_weight = 0.

    def update_values(self, **kwargs):

        # keep track of exactly how many walker segments are run, this
        # is useful for rate calculations via Hill's relation.
        self.total_n_walker_segments += len(kwargs['new_walkers'])

        # just create the bare warp records, since we know no more
        # domain knowledge, feel free to override and add more data to
        # this table
        for warp_record in kwargs['warp_data']:

            # the cycle
            cycle_idx = kwargs['cycle_idx']

            # the individual values from the warp record
            weight = warp_record['weight'][0]
            walker_idx = warp_record['walker_idx'][0]
            target_idx = warp_record['target_idx'][0]

            # determine if it was discontinuous

            # all targets are discontinuous
            if self.bc_discontinuities is Ellipsis:
                discont = True
            # none of them are discontinuous
            elif self.bc_discontinuities is None:
                discont = False
            # then it is a list of the discontinuous targets
            else:
                discont = True if target_idx in self.bc_discontinuities else False

            record = (cycle_idx, walker_idx, weight, target_idx, discont)
            self.warp_records.append(record)


    def gen_fields(self, **kwargs):

        # make the table for the collected warping records
        warp_table_df = pd.DataFrame(self.warp_records, columns=self.WARP_RECORD_COLNAMES)
        warp_table_str = tabulate(warp_table_df,
                                  headers=warp_table_df.columns,
                                  tablefmt='orgtbl')

        fields = {
            'name' : self.bc_name,
            'total_n_walker_segments' : self.total_n_walker_segments,
            'total_crossings' : self.total_crossings,
            'total_crossed_weight' : self.total_crossed_weight,
            'warping_log' : warp_table_str,
        }

        return fields

    def gen_bc_section(self, **kwargs):

        section_kwargs = self.gen_fields(**kwargs)

        section_str = Template(self.BC_SECTION_TEMPLATE).render(
            **section_kwargs
        )

        return section_str





"""Class for generating layouts for resampling trees.

Routines
--------

ResamplingTreeLayout.layout

"""

from warnings import warn
from collections import defaultdict
from copy import copy
import itertools as it

import numpy as np
import networkx as nx

from wepy.analysis.network_layouts.layout import LayoutError

class ResamplingTreeLayout():
    """Class that wraps the parameters for generating resampling tree layouts.

    Use the 'layout' method to generate inputs to a LayoutGraph for rendering.

    Attributes
    ----------

    node_radius : float
        Default node radius to use.
    row_spacing : float
        Spacing between nodes in a single row in layout.
    step_spacing : float
        Spacing between the rows of nodes in each step of the layout.
    central_axis : float
        X coordinate value to center each row around in the tree layout.
    """

    def __init__(self,
                 node_radius=1.0,
                 row_spacing=5.0,
                 step_spacing=20.0,
                 central_axis=0.0):
        """Constructing the object is just a setting of the parameters and
        collection of methods for generating layout positions.

        Arguments
        ---------
        node_radius : float, optional
            Default node radius to use.
             (Default value = 1.0)

        row_spacing : float
            Spacing between nodes in a single row in layout.
             (Default value = 5.0)

        step_spacing : float
            Spacing between the rows of nodes in each step of the layout.
             (Default value = 20.0)

        central_axis : float
            X coordinate value to center each row around in the tree layout.
             (Default value = 0.0)
        """


        self.node_radius = node_radius
        self.row_spacing = row_spacing
        self.step_spacing = step_spacing
        self.central_axis = central_axis

    def _overlaps(self, positions, node_radii, node_idx):
        """

        Parameters
        ----------
        positions :

        node_radii :

        node_idx :


        Returns
        -------

        """

        # get the nodes that this one overlaps with
        overlaps = []
        for other_node_idx, other_node_position in enumerate(positions):

            # you can't overlap yourself
            if node_idx == other_node_idx:
                overlaps.append(False)
            else:
                # check if there is an overlap between nodes
                diff = np.abs(positions[node_idx] - other_node_position)
                if diff < node_radii[node_idx] + node_radii[other_node_idx] + self.row_spacing:
                    overlaps.append(True)
                else:
                    overlaps.append(False)

        return overlaps

    def _node_row_length(self, node_positions, node_radii):
        """Get the edge to edge length of a row of nodes.

        Parameters
        ----------
        node_positions :

        node_radii :

        Returns
        -------

        row_length : float

        """

        max_child_idx = np.argmax(node_positions)
        max_edge = node_positions[max_child_idx] + node_radii[max_child_idx]
        min_child_idx = np.argmin(node_positions)
        min_edge = node_positions[min_child_idx] - node_radii[min_child_idx]

        return abs(max_edge - min_edge)

    def _simple_gen_distribution(self, nodes_x, node_radii):
        """

        Parameters
        ----------
        nodes_x :

        node_radii :


        Returns
        -------

        new_nodes_positions

        """

        # we want to update the positions given so we copy that array
        node_positions = np.array(copy(nodes_x))

        # first we check to see if there are any groups of nodes that
        # are in identical positions. If there are we can lump them
        # into a single node with a bigger radii and move things
        # around it

        groups = []
        group_positions = []
        for node_idx, node_position in enumerate(node_positions):

            # first get if this nodes position is already in the list
            # of known positions
            if node_position in group_positions:
                # if it is then get which group it is and add it to the list for that
                idx = group_positions.index(node_position)
                groups[idx].append(node_idx)

            else:
                # start a new group for it
                group_positions.append(node_position)
                groups.append([node_idx])

        # now we filter out the groups that only have one node
        chosen_idxs = [idx for idx, group in enumerate(groups)
                       if len(group) > 1]

        group_positions = [group_positions[idx] for idx in chosen_idxs]

        groups = [groups[idx] for idx in chosen_idxs]
        n_groups = len(groups)

        # we make a collection of the nodes that are in groups
        grouped_node_idxs = set(it.chain(*groups))

        # then we remove these nodes from the positions and radii
        # arrays and replace them with single nodes of updated radii
        # and positions.
        group_radii = []
        for group_idx, group in enumerate(groups):

            # the new radii will be the sum of the diameters plus the
            # spacing between the nodes

            # the sum of the diamters
            sum_diameters = sum([2 * node_radii[node_idx]
                                 for node_idx in group])

            # the spacing between is for 1 less than the total number
            # of the nodes in the group
            sum_spacing = self.row_spacing * (len(group) - 1)

            # the whole groups radius
            group_radius = (sum_diameters + sum_spacing) / 2

            group_radii.append(group_radius)

        # then we rip out the nodes that are part of groups and are
        # left with the singletons
        singleton_idxs = [node_idx for node_idx, _ in enumerate(node_positions)
                          if node_idx not in grouped_node_idxs]
        singleton_positions = [node_positions[node_idx] for node_idx in singleton_idxs]
        singleton_radii = [node_radii[node_idx] for node_idx in singleton_idxs]

        # then we make the effective nodes to position later, we put
        # them at the beginning of the new positions list so we can
        # easily decompose it later and put them back out to how they
        # should be
        eff_positions = np.array(group_positions + singleton_positions)
        eff_radii = np.array(group_radii + singleton_radii)

        # now that we have made the effective nodes we go ahead and
        # place them

        # then for each node we place it sequentially
        for node_idx, node_radius in enumerate(eff_radii):

            # get the overlaps for this node
            overlaps = self._overlaps(eff_positions, eff_radii, node_idx)

            # get how many nodes overlap it to the left and to the
            # right of it's starting position
            left_overlaps = [i for i, _ in enumerate(overlaps)
                             if eff_positions[i] < eff_positions[node_idx]]
            right_overlaps = [i for i, _ in enumerate(overlaps)
                              if eff_positions[i] > eff_positions[node_idx]]

            left_n_overlaps = len(left_overlaps)
            right_n_overlaps = len(right_overlaps)

            # we just move all the other nodes to allow this node to
            # be placed where we proposed it

            # left overlaps
            if left_n_overlaps > 0:

                # repel the other nodes around it

                # get the node idxs to the left of this one
                detangle_left_node_idxs = [i for i, position in enumerate(eff_positions)
                                            if position < eff_positions[node_idx]]

                # get the (center to center) distances from this node to
                # all it's overlaps
                cc_left_overlap_dists = [abs(eff_positions[node_idx] - eff_positions[i])
                                         for i in left_overlaps]

                # then get the edge to edge distances (which can be
                # negative where they cross over) this is the sum of the
                # radii subtracted from the center to center distance (ee = d-(r_0 + r_1))
                ee_left_overlap_dists = [cc_left_overlap_dists[i] - (node_radius + eff_radii[node_i])
                                         for i, node_i in enumerate(left_overlaps)]

                # we want the minimum one of the edge to edges (even if
                # its negative)
                left_min_ee_overlap_dist = min(ee_left_overlap_dists)

                # this overlap should always be negative for the
                # overlapping nodes so we minimally need to move one
                # node this far away to have no overlap, on top of
                # that we add the spacing for the move distance to make
                left_move = -1 * (self.row_spacing + (-left_min_ee_overlap_dist))

                # then apply to those nodes
                eff_positions[detangle_left_node_idxs] += left_move


            # right overlaps
            if right_n_overlaps > 0:

                # see above for comments
                detangle_right_node_idxs = [i for i, position in enumerate(eff_positions)
                                            if position > eff_positions[node_idx]]

                cc_right_overlap_dists = [abs(eff_positions[node_idx] - eff_positions[i])
                                          for i in right_overlaps]

                ee_right_overlap_dists = [cc_right_overlap_dists[i] - (node_radius + eff_radii[node_i])
                                         for i, node_i in enumerate(right_overlaps)]

                right_min_ee_overlap_dist = min(ee_right_overlap_dists)

                right_move = +1 * (self.row_spacing + (-right_min_ee_overlap_dist))

                eff_positions[detangle_right_node_idxs] += right_move


        # now that we have placed the effective nodes we want to
        # reconstruct the original nodes from the grouped ones

        # make a new positions array that we can repopulate
        new_node_positions = [None for _ in node_positions]

        # first for each group we need to place the nodes that had
        # identical positions within the space created by their larger
        # effective nodes
        for group_idx, group in enumerate(groups):

            # the position of the group overall, since the grouped
            # nodes were at the front of the list of the effective
            # nodes with preserverd order we can use the group_index
            # to get them
            group_position = eff_positions[group_idx]
            group_radii = eff_radii[group_idx]

            member_radii = [node_radii[node_idx] for node_idx in group]

            # just place them like you would the initial distribution,
            # which is centered at 0
            member_positions = np.array(self._initial_parent_distribution(member_radii))

            # translate all nodes left (negative by the midpoint/radius)
            member_positions -= group_radii

            # then we translate all of them to the overall group position
            member_positions += group_position

            # now that we have good positions we can place them back
            # into the main level positions list
            for i, node_idx in enumerate(group):
                new_node_positions[node_idx] = member_positions[i]

        # The singletons were at the end of the effective positions so
        # we have N_group_nodes + i where i is the index over the singletons
        for i, node_idx in enumerate(singleton_idxs):

            new_node_positions[node_idx] = eff_positions[n_groups + i]

        # sanity check that we covered them all
        assert all([True if pos is not None else False
                    for pos in new_node_positions]),\
                        "not all positions recovered from the effective nodes"


        return new_node_positions

    def _simple_next_gen(self, parents_x, children_parent_idxs, node_radii):
        """

        Parameters
        ----------
        parents_x :

        children_parent_idxs :

        node_radii :


        Returns
        -------

        children_x

        """

        children_x = []
        for parent_idx in children_parent_idxs:
            children_x.append(parents_x[parent_idx])

        children_x = self._simple_gen_distribution(children_x, node_radii)

        # check to make sure there are no overlaps
        for node_idx in range(len(children_x)):
            overlaps = self._overlaps(children_x, node_radii, node_idx)
            if any(overlaps):
                pass
                #raise LayoutError("node {} has an overlap".format(node_idx))

        return children_x


    def _initial_parent_distribution(self, node_radii):
        """

        Parameters
        ----------
        node_radii :

        Returns
        -------

        positions

        """

        # start at the origin and add nodes at positions that
        # accomodate their diameter and that are spaced by spacing_factor
        origin = 0.0

        positions = []

        # place the first node at the origin, offset by its radius so
        # that the leading edge is at 0.0, then iterate for the rest
        # of them
        positions.append(origin + node_radii[0])

        # compute the edge so we know how much to space the next one
        # by
        last_edge = origin + 2 * node_radii[0]

        # then do the rest
        for node_idx, node_radius in enumerate(node_radii[1:]):

            node_idx += 1

            # set the position from the last edge plus the space and
            # the new nodes radius
            position = last_edge + self.row_spacing + node_radius

            positions.append(position)

            # then calculate that ones edge
            last_edge = position + node_radius

        return positions

    def _center_row(self, positions, radii, center):
        """Centers the midpoint of a row of nodes around a number.

        Parameters
        ----------
        positions :

        radii :

        center : float


        Returns
        -------

        centered_positions

        """

        # calculate the length of the row
        row_length = self._node_row_length(positions, radii)

        min_child_idx = np.argmin(positions)
        min_edge = positions[min_child_idx] - radii[min_child_idx]

        # row midpoint is the minimum edge plus the radius of the row
        row_midpoint = min_edge + (row_length / 2)

        # translate all the points so that the midpoint is at 0
        centered_positions = positions - row_midpoint

        return centered_positions

    def _layout_array(self, parent_table, radii_array):
        """Generates the structured array of positions for nodes in the parent
        forest tree layout given the corresponding radii.

        Parameters
        ----------
        parent_table :

        radii_array :

        Returns
        -------

        node_positions

        """

        n_timesteps = len(parent_table)
        n_walkers = len(parent_table[0])

        # initialize the positions to zeros, we add one to the
        # timesteps for the roots of the resampling trees
        node_positions = np.zeros((n_timesteps+1, n_walkers, 3))

        # initialize the first generation (cycle) node positions, this
        # is the root positions, we give it the radii of the first row in the array
        first_gen_positions = np.array(self._initial_parent_distribution(radii_array[0]))

        # then center them around the central axis
        first_gen_positions = self._center_row(first_gen_positions, radii_array[0],
                                               self.central_axis)

        # save them as full coordinates for visualization
        node_positions[0] = np.array([[x, 0.0, 0.0] for x in first_gen_positions])

        # we use the last gen positions to make the next gen
        last_gen_positions = first_gen_positions

        # layout all the rest of the steps
        for step_idx in range(n_timesteps):

            # the generation index is the one used for the actual tree
            # and it is always one more than the step index, since the
            # first generation is not counted as a step and was
            # already made outside the loop
            generation_idx = step_idx + 1

            # generate the starting positions for the next generation nodes
            curr_gen_positions = self._simple_next_gen(last_gen_positions,
                                                       parent_table[step_idx],
                                                       radii_array[generation_idx])

            # center them around the central axis
            curr_gen_positions = self._center_row(curr_gen_positions,
                                                  radii_array[generation_idx],
                                                   self.central_axis)


            # figure out how big the increase in the Y direction
            # should be. This is based on the largest radii of the
            # last step and the largets radii of this step with the
            # mandatory spacing in between

            # get the y dimension of the last step
            last_y = node_positions[generation_idx-1, 0, 1]

            # get the largest radii of the last step
            last_max_radius = max(radii_array[generation_idx-1])

            # get the largest radii of this step
            this_max_radius = max(radii_array[generation_idx])

            # compute the step y dimension, which is from the last y
            # value plus the max radii plus the spacing
            step_y = last_y + last_max_radius + self.step_spacing + this_max_radius

            # then generate the coordinates
            node_positions[generation_idx] = np.array(
                [np.array([x, step_y, 0.0])
                 for x in curr_gen_positions])

            # set the last gen positions
            last_gen_positions = curr_gen_positions

        return node_positions


    def layout(self, parent_forest, node_radii=None):
        """Given a parent forest object, returns a dictionary
        mapping nodes to their xyz layout coordinates.

        If the node radii are given (as a dictionary mapping node ID
        to the desired radius) these are used as the the node radii
        and the default node radius parameter of the layout object is
        ignored.

        Radii are needed to calculate the proper positions so there is
        no overlaps between nodes.

        Parameters
        ----------
        parent_forest : ParentForest object

        node_radii : dict of node_id: float
            A dictionary mapping the nodes to node radii.
             (Default value = None)


        Returns
        -------
        node_coords : dict of node_id: array_like of float of dim (3,)
            x, y, z coordinates for all nodes. Z will be 0 for all.

        Raises
        ------
        ValueError
            If an invalid node_id is given.

        Warns
        -----
        If some but not all nodes were given assigned radii. Uses
        default value for unspecified nodes.

        """

        # get the parent table from the parent forest
        parent_table = parent_forest.parent_table

        # default your node radii to the layout parameter, and fill in
        # the rest with the node_radii given to this method

        # we need to generate an array of node radii, the size of
        # which is the parent table plus one extra row for the
        # root parents
        radii_array = []

        # count the number of nodes to check later whether or not the
        # user has given complete coverage with their radii. We can
        # issue a warning so that the user is not confused.
        n_nodes = 0

        # the "root" row, it is the same length as the first in the
        # parent table
        radii_array.append([self.node_radius for col in parent_table[0]])
        n_nodes += len(parent_table[0])

        # the rest of them
        for row in parent_table:
            radii_array.append([self.node_radius for col in row])
            n_nodes += len(row)

        # if any radii are given overwrite the default size, we count
        # how many new radii we give so that we can compare to the
        # actual number of nodes
        n_new_radii = 0
        if node_radii is not None:
            for node_id, radius in node_radii.items():

                # we know that the node_id is the (cycle_idx,
                # walker_idx), so we use these as indices on the array
                cycle_idx, walker_idx = node_id

                # the indices in the node_array start at the -1 root
                # nodes so we increase cycle index by one to match,
                # this is called the generation index
                generation_idx = cycle_idx + 1

                try:
                    radii_array[generation_idx][walker_idx] = radius
                except IndexError:
                    raise ValueError("invalid node_id was given")

                n_new_radii += 1

        # check if we have fewer assignments than there were nodes
        if n_new_radii < n_nodes:
            warn(
                "not all nodes were assigned custom radii,"
                " default value {} was used instead".format(self.node_radius))

        layout_array = self._layout_array(parent_table, radii_array)

        # make a dictionary mapping node ids to layout values
        node_coords = {}
        for cycle_idx, layout_row in enumerate(layout_array):

            for walker_idx, coord in enumerate(layout_row):

                # since the layout array starts at the root nodes (cycle
                # -1) we just reduce the cycle_idx by one when we set the
                # node attribute
                node_coords[(cycle_idx-1, walker_idx)] = coord

        return node_coords
